import csv
from pathlib import Path
import sys
from time import time

import numpy as np

from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, save_nifti
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import load_tractogram, save_tractogram
from dipy.nn.synthseg import SynthSeg
from dipy.nn.torch.synthseg import have_torch
from dipy.segment.bundles import RecoBundles
from dipy.segment.clustering import QuickBundles, qbx_and_merge
from dipy.segment.fss import FastStreamlineSearch, nearest_from_matrix_col
from dipy.segment.mask import median_otsu
from dipy.segment.tissue import TissueClassifierHMRF, dam_classifier
from dipy.tracking import Streamlines
from dipy.utils.deprecator import deprecated_params
from dipy.utils.logging import logger
from dipy.workflows.base import format_key_value_table
from dipy.workflows.utils import handle_vol_idx
from dipy.workflows.workflow import Workflow


class MedianOtsuFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "medotsu"

    @deprecated_params(["autocrop"], since="1.11.0", until="1.13.0")
    def run(
        self,
        input_files,
        bvalues_files=None,
        b0_threshold=50,
        save_masked=False,
        median_radius=2,
        numpass=5,
        autocrop=False,
        vol_idx=None,
        dilate=None,
        finalize_mask=False,
        out_dir="",
        out_mask="brain_mask.nii.gz",
        out_masked="dwi_masked.nii.gz",
    ):
        """Workflow wrapping the median_otsu segmentation method.

        Applies median_otsu segmentation on each file found by 'globing'
        ``input_files`` and saves the results in a directory specified by
        ``out_dir``.

        Parameters
        ----------
        input_files : string or Path
            Path to the input volumes. This path may contain wildcards to
            process multiple inputs at once.
        bvalues_files : variable string or Path, optional
            Path to the b-values files. This path may contain wildcards to
            process multiple inputs at once.
        b0_threshold : float, optional
            Threshold to consider a volume as b0. Used only if bvalues_files
            is provided.
        save_masked : bool, optional
            Save mask.
        median_radius : int, optional
            Radius (in voxels) of the applied median filter.
        numpass : int, optional
            Number of pass of the median filter.
        autocrop : bool, optional
            .. deprecated:: 1.11.0
               This parameter is deprecated and will be removed in 1.13.0.

            If True, the mask is cropped to the bounding box of the
            non-zero elements. This is useful for large 3D volumes
            (e.g. 4D volumes) where the mask is not the same size as
            the input volume.
        vol_idx : str, optional
            1D array representing indices of ``axis=-1`` of a 4D
            `input_volume`. From the command line use something like
            '1,2,3-5,7'. This input is required for 4D volumes if bval files
            are not provided. If bval files are provided, vol_idx is
            ignored and b0 volumes are used for mask computation.
        dilate : int, optional
            number of iterations for binary dilation.
        finalize_mask : bool, optional
            Whether to remove potential holes or islands.
            Useful for solving minor errors.
        out_dir : string or Path, optional
            Output directory.
        out_mask : string, optional
            Name of the mask volume to be saved.
        out_masked : string, optional
            Name of the masked volume to be saved.
        """
        io_it = self.get_io_iterator()

        if vol_idx is not None and bvalues_files is not None and io_it:
            logger.warning(
                "'vol_idx' parameter is ignored when 'bvalues_files' is provided."
            )

        if bvalues_files is not None and not isinstance(bvalues_files, list):
            bvalues_files = [bvalues_files]

        if len(bvalues_files or []) > 0 and io_it:
            if len(bvalues_files) != len(io_it.inputs):
                logger.error(
                    "Number of b-values files must match the number of "
                    "input volumes."
                )
                sys.exit(1)

        vol_idx = handle_vol_idx(vol_idx)

        bvals_counter = 0
        for fpath, mask_out_path, masked_out_path in io_it:
            logger.info(f"Applying median_otsu segmentation on {fpath}")

            data, affine, img = load_nifti(fpath, return_img=True)
            vol_idx_used = vol_idx
            if bvalues_files is not None:
                bvals, _ = read_bvals_bvecs(bvalues_files[bvals_counter], None)
                bvals_counter += 1
                b0_indices = np.where(bvals <= b0_threshold)[0]
                vol_idx_used = b0_indices

            logger.debug(f"vol_idx_used: {vol_idx_used}")
            extra_args = {} if not autocrop else {"autocrop": autocrop}
            masked_volume, mask_volume = median_otsu(
                data,
                vol_idx=vol_idx_used,
                median_radius=median_radius,
                numpass=numpass,
                dilate=dilate,
                finalize_mask=finalize_mask,
                **extra_args,
            )

            save_nifti(mask_out_path, mask_volume.astype(np.float64), affine)

            logger.info(f"Mask saved as {mask_out_path}")

            if save_masked:
                save_nifti(masked_out_path, masked_volume, affine, hdr=img.header)

                logger.info(f"Masked volume saved as {masked_out_path}")

        return io_it


class RecoBundlesFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "recobundles"

    def run(
        self,
        streamline_files,
        model_bundle_files,
        greater_than=50,
        less_than=1000000,
        no_slr=False,
        clust_thr=15.0,
        reduction_thr=15.0,
        reduction_distance="mdf",
        model_clust_thr=2.5,
        pruning_thr=8.0,
        pruning_distance="mdf",
        slr_metric="symmetric",
        slr_transform="similarity",
        slr_matrix="small",
        refine=False,
        r_reduction_thr=12.0,
        r_pruning_thr=6.0,
        no_r_slr=False,
        out_dir="",
        out_recognized_transf="recognized.trx",
        out_recognized_labels="labels.npy",
    ):
        """Recognize bundles

        See :footcite:p:`Garyfallidis2018` and :footcite:p:`Chandio2020a` for
        further details about the method.

        Parameters
        ----------
        streamline_files : string or Path
            The path of streamline files where you want to recognize bundles.
        model_bundle_files : string or Path
            The path of model bundle files.
        greater_than : int, optional
            Keep streamlines that have length greater than
            this value in mm.
        less_than : int, optional
            Keep streamlines have length less than this value
            in mm.
        no_slr : bool, optional
            Don't enable local Streamline-based Linear
            Registration.
        clust_thr : float, optional
            MDF distance threshold for all streamlines.
        reduction_thr : float, optional
            Reduce search space by (mm).
        reduction_distance : string, optional
            Reduction distance type can be mdf or mam.
        model_clust_thr : float, optional
            MDF distance threshold for the model bundles.
        pruning_thr : float, optional
            Pruning after matching.
        pruning_distance : string, optional
            Pruning distance type can be mdf or mam.
        slr_metric : string, optional
            Options are None, symmetric, asymmetric or diagonal.
        slr_transform : string, optional
            Transformation allowed. translation, rigid, similarity or scaling.
        slr_matrix : string, optional
            Options are 'nano', 'tiny', 'small', 'medium', 'large', 'huge'.
        refine : bool, optional
            Enable refine recognized bundle.
        r_reduction_thr : float, optional
            Refine reduce search space by (mm).
        r_pruning_thr : float, optional
            Refine pruning after matching.
        no_r_slr : bool, optional
            Don't enable Refine local Streamline-based Linear
            Registration.
        out_dir : string or Path, optional
            Output directory.
        out_recognized_transf : string, optional
            Recognized bundle in the space of the model bundle.
        out_recognized_labels : string, optional
            Indices of recognized bundle in the original tractogram.

        References
        ----------
        .. footbibliography::

        """
        slr = not no_slr
        r_slr = not no_r_slr

        bounds = [
            (-30, 30),
            (-30, 30),
            (-30, 30),
            (-45, 45),
            (-45, 45),
            (-45, 45),
            (0.8, 1.2),
            (0.8, 1.2),
            (0.8, 1.2),
        ]

        slr_matrix = slr_matrix.lower()
        if slr_matrix == "nano":
            slr_select = (100, 100)
        if slr_matrix == "tiny":
            slr_select = (250, 250)
        if slr_matrix == "small":
            slr_select = (400, 400)
        if slr_matrix == "medium":
            slr_select = (600, 600)
        if slr_matrix == "large":
            slr_select = (800, 800)
        if slr_matrix == "huge":
            slr_select = (1200, 1200)

        slr_transform = slr_transform.lower()
        if slr_transform == "translation":
            bounds = bounds[:3]
        if slr_transform == "rigid":
            bounds = bounds[:6]
        if slr_transform == "similarity":
            bounds = bounds[:7]
        if slr_transform == "scaling":
            bounds = bounds[:9]

        logger.info("### RecoBundles ###")

        io_it = self.get_io_iterator()

        t = time()
        logger.info(streamline_files)
        input_obj = load_tractogram(streamline_files, "same", bbox_valid_check=False)
        streamlines = input_obj.streamlines

        logger.info(f" Loading time {time() - t:0.3f} sec")

        rb = RecoBundles(streamlines, greater_than=greater_than, less_than=less_than)
        bundle_summary = {}

        for _, mb, out_rec, out_labels in io_it:
            bundle_name = Path(mb).stem
            t = time()
            logger.info(mb)
            model_bundle = load_tractogram(
                mb, "same", bbox_valid_check=False
            ).streamlines
            logger.info(f" Loading time {time() - t:0.3f} sec")
            logger.info("model file = ")
            logger.info(mb)

            recognized_bundle, labels = rb.recognize(
                model_bundle,
                model_clust_thr=model_clust_thr,
                reduction_thr=reduction_thr,
                reduction_distance=reduction_distance,
                pruning_thr=pruning_thr,
                pruning_distance=pruning_distance,
                slr=slr,
                slr_metric=slr_metric,
                slr_x0=slr_transform,
                slr_bounds=bounds,
                slr_select=slr_select,
                slr_method="L-BFGS-B",
            )

            if refine:
                if len(recognized_bundle) > 1:
                    # affine
                    x0 = np.array([0, 0, 0, 0, 0, 0, 1.0, 1.0, 1, 0, 0, 0])
                    affine_bounds = [
                        (-30, 30),
                        (-30, 30),
                        (-30, 30),
                        (-45, 45),
                        (-45, 45),
                        (-45, 45),
                        (0.8, 1.2),
                        (0.8, 1.2),
                        (0.8, 1.2),
                        (-10, 10),
                        (-10, 10),
                        (-10, 10),
                    ]

                    recognized_bundle, labels = rb.refine(
                        model_bundle,
                        recognized_bundle,
                        model_clust_thr=model_clust_thr,
                        reduction_thr=r_reduction_thr,
                        reduction_distance=reduction_distance,
                        pruning_thr=r_pruning_thr,
                        pruning_distance=pruning_distance,
                        slr=r_slr,
                        slr_metric=slr_metric,
                        slr_x0=x0,
                        slr_bounds=affine_bounds,
                        slr_select=slr_select,
                        slr_method="L-BFGS-B",
                    )

            if len(labels) > 0:
                ba, bmd = rb.evaluate_results(
                    model_bundle, recognized_bundle, slr_select
                )

                logger.info(f"Bundle adjacency Metric {ba}")
                logger.info(f"Bundle Min Distance Metric {bmd}")

            bundle_summary[bundle_name] = len(labels)
            new_tractogram = StatefulTractogram(
                recognized_bundle, streamline_files, Space.RASMM
            )
            save_tractogram(new_tractogram, out_rec, bbox_valid_check=False)
            logger.info("Saving output files ...")
            np.save(out_labels, np.array(labels))
            logger.info(out_rec)
            logger.info(out_labels)

        summary_str = {name: str(count) for name, count in bundle_summary.items()}
        table = format_key_value_table(
            summary_str, "Bundle", "# Streamlines", sort=False
        )
        logger.info("\nBundle Recognition Summary:\n" + table)


class LabelsBundlesFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "labelsbundles"

    def run(
        self,
        streamline_files,
        labels_files,
        out_dir="",
        out_bundle="recognized_orig.trx",
    ):
        """Extract bundles using existing indices (labels)

        See :footcite:p:`Garyfallidis2018` for further details about the method.

        Parameters
        ----------
        streamline_files : string or Path
            The path of streamline files where you want to recognize bundles.
        labels_files : string or Path
            The path of model bundle files.
        out_dir : string or Path, optional
            Output directory.
        out_bundle : string, optional
            Recognized bundle in the space of the model bundle.

        References
        ----------
        .. footbibliography::

        """
        logger.info("### Labels to Bundles ###")

        io_it = self.get_io_iterator()
        for f_steamlines, f_labels, out_bundle in io_it:
            logger.info(f_steamlines)
            sft = load_tractogram(f_steamlines, "same", bbox_valid_check=False)
            streamlines = sft.streamlines

            logger.info(f_labels)
            location = np.load(f_labels)
            if len(location) < 1:
                bundle = Streamlines([])
            else:
                bundle = streamlines[location]

            logger.info("Saving output files ...")
            new_sft = StatefulTractogram(bundle, sft, Space.RASMM)
            save_tractogram(new_sft, out_bundle, bbox_valid_check=False)
            logger.info(out_bundle)


class ClassifyTissueFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "extracttissue"

    def run(
        self,
        input_files,
        bvals_file=None,
        method=None,
        wm_threshold=0.5,
        b0_threshold=50,
        low_signal_threshold=50,
        nclass=None,
        beta=0.1,
        tolerance=1e-05,
        max_iter=100,
        out_dir="",
        out_tissue="tissue_classified.nii.gz",
        out_pve="tissue_classified_pve.nii.gz",
        out_csv="tissue_classified.csv",
    ):
        """Extract tissue from a volume.

        Parameters
        ----------
        input_files : string or Path
            Path to the input volumes. This path may contain wildcards to
            process multiple inputs at once.
        bvals_file : string or Path, optional
            Path to the b-values file. Required for 'dam' method.
        method : string, optional
            Method to use for tissue extraction. Options are:

            - 'hmrf': Markov Random Fields modeling approach.
            - 'dam': Directional Average Maps, proposed by
              :footcite:p:`Cheng2020`.
            - 'synthseg': SynthSeg method based on Freesurfer's implementation,
              proposed by :footcite:p:`Billot2023`.

            'hmrf' method is recommended for T1w images, while 'dam' method is
            recommended for DWI Multishell images (single shell are not recommended).
            'synthseg' works for all modalities, but requires more memory.
            No pve maps are generated for 'synthseg' method.
        wm_threshold : float, optional
            The threshold below which a voxel is considered white matter. For data like
            HCP, threshold of 0.5 proves to be a good choice. For data like cfin, higher
            threshold values like 0.7 or 0.8 are more suitable. Used for 'dam' method.
        b0_threshold : float, optional
            The intensity threshold for a b=0 image. used only for 'dam' method.
        low_signal_threshold : float, optional
            The threshold below which a voxel is considered to have low signal.
            Used only for 'dam' method.
        nclass : int, optional
            Number of desired classes. Used only for 'hmrf' method.
        beta : float, optional
            Smoothing parameter, the higher this number the smoother the
            output will be. Used only for 'hmrf' method.
        tolerance : float, optional
            Value that defines the percentage of change tolerated to
            prevent the ICM loop to stop. Default is 1e-05.
            If you want tolerance check to be disabled put 'tolerance = 0'.
            Used only for 'hmrf' method.
        max_iter : int, optional
            Fixed number of desired iterations. Default is 100.
            This parameter defines the maximum number of iterations the
            algorithm will perform. The loop may terminate early if the
            change in energy sum between iterations falls below the
            threshold defined by `tolerance`. However, if `tolerance` is
            explicitly set to 0, this early stopping mechanism is disabled,
            and the algorithm will run for the specified number of
            iterations unless another stopping criterion is met.
            Used only for 'hmrf' method.
        out_dir : string or Path, optional
            Output directory.
        out_tissue : string, optional
            Name of the tissue volume to be saved.
        out_pve : string, optional
            Name of the pve volume to be saved.
        out_csv : string, optional
            Name of the csv file containing label names to be saved.

        REFERENCES
        ----------
        .. footbibliography::

        """
        csv_fieldnames = ["Label", "Name"]
        io_it = self.get_io_iterator()

        if not method or method.lower() not in ["hmrf", "dam", "synthseg"]:
            logger.error(
                f"Unknown method '{method}' for tissue extraction. "
                "Choose '--method hmrf' (for T1w) or '--method dam' (for DWI)"
                " or '--method synthseg' (for all modalities)."
            )
            sys.exit(1)

        for fpath, tissue_out_path, opve, ocsv in io_it:
            logger.info(f"Extracting tissue from {fpath}")

            data, affine = load_nifti(fpath)
            class_list = []

            if method.lower() == "hmrf":
                if nclass is None:
                    logger.error(
                        "Number of classes is required for 'hmrf' method. "
                        "For example, Use '--nclass 4' to specify the number of "
                        "classes."
                    )
                    sys.exit(1)
                classifier = TissueClassifierHMRF()
                _, segmentation_final, PVE = classifier.classify(
                    data, nclass, beta, tolerance=tolerance, max_iter=max_iter
                )

                save_nifti(tissue_out_path, segmentation_final, affine)
                save_nifti(opve, PVE, affine)
                class_list.append(["0", "Background"])
                for i in range(1, nclass + 1):
                    class_list.append([f"{i}", f"Tissue_{i}"])
                class_list.append(
                    [
                        "# Due to the nature of the HMRF algorithm",
                        "tissues are not labeled specifically.",
                    ]
                )

            elif method.lower() == "dam":
                if bvals_file is None or not Path(bvals_file).is_file():
                    logger.error("'--bvals filename' is required for 'dam' method")
                    sys.exit(1)

                bvals, _ = read_bvals_bvecs(bvals_file, None)
                wm_mask, gm_mask = dam_classifier(
                    data,
                    bvals,
                    wm_threshold=wm_threshold,
                    b0_threshold=b0_threshold,
                    low_signal_threshold=low_signal_threshold,
                )
                result = np.zeros(wm_mask.shape, dtype=np.int32)
                result[wm_mask] = 1
                result[gm_mask] = 2
                save_nifti(tissue_out_path, result, affine)
                save_nifti(
                    opve, np.stack([wm_mask, gm_mask], axis=-1).astype(np.int32), affine
                )
                class_list.append(["0", "Background"])
                class_list.append(["1", "White_Matter"])
                class_list.append(["2", "Gray_Matter"])

            elif method.lower() == "synthseg":
                synthseg = SynthSeg(verbose=True)
                segmentation_final, label_dict, _ = synthseg.predict(data, affine)
                for label, name in label_dict.items():
                    class_list.append([f"{label}", name])

                save_nifti(tissue_out_path, segmentation_final.astype(np.int32), affine)

            with open(ocsv, "w", newline="") as csv_object:
                csv_writer = csv.writer(csv_object)
                csv_writer.writerow(csv_fieldnames)
                csv_writer.writerows(class_list)

            if not method.lower() == "synthseg":
                logger.info(f"Tissue saved as {tissue_out_path} and PVE as {opve}")
            else:
                logger.info(f"Tissue saved as {tissue_out_path}")
            logger.info(f"Label names saved as {ocsv}")

        return io_it


class BrainMaskFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "brainmask"

    def run(
        self,
        input_files,
        *,
        method="synthseg",
        bvalues_files=None,
        b0_threshold=50,
        median_radius=2,
        numpass=5,
        vol_idx=None,
        dilate=None,
        finalize_mask=False,
        save_masked=False,
        use_cuda=False,
        out_dir="",
        out_mask="brain_mask.nii.gz",
        out_masked="brain_masked.nii.gz",
    ):
        """Unified brain masking workflow supporting multiple methods.

        Applies brain masking on each file found by 'globing' ``input_files``
        and saves the results in a directory specified by ``out_dir``.

        Parameters
        ----------
        input_files : string or Path
            Path to the input volumes. This path may contain wildcards to
            process multiple inputs at once.
        method : string, optional
            Brain masking method. Options are:

            - 'median_otsu': Classical median Otsu segmentation, DWI-focused.
            - 'evac': Deep learning EVACPlus method, T1-focused.
            - 'synthseg': Deep learning SynthSeg method, works for all
              modalities.

        bvalues_files : variable string or Path, optional
            Path to the b-values files. Used only for 'median_otsu' method.
            This path may contain wildcards to process multiple inputs at once.
        b0_threshold : float, optional
            Threshold to consider a volume as b0. Used only if bvalues_files
            is provided with 'median_otsu' method.
        median_radius : int, optional
            Radius (in voxels) of the applied median filter. Used only for
            'median_otsu' method.
        numpass : int, optional
            Number of pass of the median filter. Used only for 'median_otsu'
            method.
        vol_idx : str, optional
            1D array representing indices of ``axis=-1`` of a 4D
            `input_volume`. From the command line use something like
            '1,2,3-5,7'. Used only for 'median_otsu' method. This input is
            required for 4D volumes if bval files are not provided. If bval
            files are provided, vol_idx is ignored and b0 volumes are used
            for mask computation.
        dilate : int, optional
            Number of iterations for binary dilation. Used only for
            'median_otsu' method.
        finalize_mask : bool, optional
            Whether to remove potential holes or islands. Useful for solving
            minor errors. Used only for 'median_otsu' method.
        save_masked : bool, optional
            Save the masked volume in addition to the mask.
        use_cuda : bool, optional
            Use CUDA for GPU acceleration. Used only for 'evac' and
            'synthseg' methods.
        out_dir : string or Path, optional
            Output directory.
        out_mask : string, optional
            Name of the mask volume to be saved.
        out_masked : string, optional
            Name of the masked volume to be saved.
        """
        valid_methods = ["median_otsu", "synthseg", "evac"]
        method = method.lower()

        if method not in valid_methods:
            logger.error(
                f"Unknown method '{method}' for brain masking. "
                f"Choose one of: {', '.join(valid_methods)}."
            )
            sys.exit(1)

        if method == "synthseg" and not have_torch:
            logger.warning(
                "SynthSeg method requires PyTorch. Please install PyTorch to use "
                "this method. Switching to 'median_otsu' method instead."
            )
            method = "median_otsu"

        logger.info(f"Brain masking using method: '{method}'")

        io_it = self.get_io_iterator()

        if method == "median_otsu":
            if vol_idx is not None and bvalues_files is not None and io_it:
                logger.warning(
                    "'vol_idx' parameter is ignored when 'bvalues_files' is "
                    "provided."
                )

            if bvalues_files is not None and not isinstance(bvalues_files, list):
                bvalues_files = [bvalues_files]

            if len(bvalues_files or []) > 0 and io_it:
                if len(bvalues_files) != len(io_it.inputs):
                    logger.error(
                        "Number of b-values files must match the number of "
                        "input volumes."
                    )
                    sys.exit(1)

            vol_idx = handle_vol_idx(vol_idx)

        elif method == "evac":
            logger.info("Loading EVACPlus model...")
            from dipy.nn.evac import EVACPlus

            evac_model = EVACPlus(use_cuda=use_cuda)
            logger.info("EVACPlus model loaded.")

        elif method == "synthseg":
            logger.info("Loading SynthSeg model...")
            synthseg_model = SynthSeg(verbose=True, use_cuda=use_cuda)
            logger.info("SynthSeg model loaded.")

        bvals_counter = 0
        for fpath, mask_out_path, masked_out_path in io_it:
            logger.info(f"Applying '{method}' brain masking on {fpath}")

            data, affine, img = load_nifti(fpath, return_img=True)
            logger.info(f"Input shape: {data.shape}")

            if method == "median_otsu":
                if not vol_idx and not bvalues_files and data.ndim == 4:
                    logger.error(
                        "Either 'vol_idx' or 'bvalues_files' must be provided "
                        "for 'median_otsu' method."
                    )
                    sys.exit(1)
                vol_idx_used = vol_idx
                if bvalues_files is not None:
                    bvals, _ = read_bvals_bvecs(bvalues_files[bvals_counter], None)
                    bvals_counter += 1
                    b0_indices = np.where(bvals <= b0_threshold)[0]
                    vol_idx_used = b0_indices

                logger.debug(f"vol_idx_used: {vol_idx_used}")
                masked_volume, mask_volume = median_otsu(
                    data,
                    vol_idx=vol_idx_used,
                    median_radius=median_radius,
                    numpass=numpass,
                    dilate=dilate,
                    finalize_mask=finalize_mask,
                )

            elif method == "evac":
                if data.ndim == 4:
                    logger.warning(
                        f"{fpath} is a 4D image. Only the first volume will "
                        "be used for 'evac' brain masking."
                    )
                vol = data[..., 0] if data.ndim == 4 else data
                mask_volume = evac_model.predict(vol, affine)
                mask_for_apply = (
                    mask_volume[..., np.newaxis] if data.ndim == 4 else mask_volume
                )
                masked_volume = data * mask_for_apply

            elif method == "synthseg":
                if data.ndim == 4:
                    logger.warning(
                        f"{fpath} is a 4D image. Only the first volume will "
                        "be used for 'synthseg' brain masking."
                    )
                vol = data[..., 0] if data.ndim == 4 else data
                _, _, mask_volume = synthseg_model.predict(vol, affine)
                mask_for_apply = (
                    mask_volume[..., np.newaxis] if data.ndim == 4 else mask_volume
                )
                masked_volume = data * mask_for_apply

            save_nifti(mask_out_path, mask_volume.astype(np.float64), affine)
            logger.info(f"Mask saved as {mask_out_path}")

            if save_masked:
                save_nifti(masked_out_path, masked_volume, affine, hdr=img.header)
                logger.info(f"Masked volume saved as {masked_out_path}")

            logger.info(f"Brain masking of {fpath} completed.")

        logger.info("Brain masking finished.")
        return io_it


class ClusterStreamlinesFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "cluster"

    def run(
        self,
        streamline_files,
        *,
        method="qbx_and_merge",
        threshold=10.0,
        thresholds="30,20,10",
        nb_pts=20,
        min_cluster_size=1,
        select_randomly=None,
        max_radius=10.0,
        out_dir="",
        out_centroids="centroids.trx",
        out_cluster_labels="cluster_labels.npy",
    ):
        """Cluster streamlines.

        Algorithms used to cluster streamlines are QuickBundles, QuickBundlesX or
        FastStreamlineSearch.

        Parameters
        ----------
        streamline_files : string or Path
            Path to the streamline files. This path may contain wildcards to
            process multiple inputs at once.
        method : string, optional
            Clustering method to use. Options are:

            - 'quickbundles': QuickBundles with a single distance threshold.
            - 'qbx_and_merge': QuickBundlesX with multi-level thresholds,
              then merge.
            - 'faststreamlines': QuickBundles for initial centroids followed
              by FastStreamlineSearch for streamline reassignment.

        threshold : float, optional
            Distance threshold (mm) used by 'quickbundles' and
            'faststreamlines' methods.
        thresholds : string, optional
            Comma-separated distance thresholds (mm) used by the
            'qbx_and_merge' method (e.g. '30,20,10').
        nb_pts : int, optional
            Number of points to resample each streamline to before
            clustering. Used by 'qbx_and_merge'.
        min_cluster_size : int, optional
            Minimum number of streamlines a cluster must contain to be kept.
        select_randomly : int, optional
            Randomly select a subset of streamlines before clustering.
            Used by 'qbx_and_merge'.
        max_radius : float, optional
            Maximum search radius (mm) used by the 'faststreamlines' method.
        out_dir : string or Path, optional
            Output directory.
        out_centroids : string, optional
            Filename for the output cluster centroids tractogram.
        out_cluster_labels : string, optional
            Filename for the output cluster label array (.npy).

        """
        valid_methods = ["quickbundles", "qbx_and_merge", "faststreamlines"]
        if method.lower() not in valid_methods:
            logger.error(
                f"Unknown method '{method}' for streamline clustering. "
                f"Choose one of: {', '.join(valid_methods)}."
            )
            sys.exit(1)

        thresholds_list = [float(t) for t in thresholds.split(",")]

        logger.info(f"Clustering streamlines using method: '{method}'")

        io_it = self.get_io_iterator()

        for fpath, centroids_out_path, labels_out_path in io_it:
            logger.info(f"Loading streamlines from {fpath}")
            sft = load_tractogram(fpath, "same", bbox_valid_check=False)
            streamlines = sft.streamlines
            logger.info(f"Loaded {len(streamlines)} streamlines")

            if method.lower() == "quickbundles":
                cluster_map = QuickBundles(threshold).cluster(streamlines)
                n_total = len(cluster_map)
                out_centroids_sl = Streamlines()
                labels = np.full(len(streamlines), -1, dtype=np.int32)
                cluster_id = 0
                for cluster in cluster_map:
                    if len(cluster) >= min_cluster_size:
                        out_centroids_sl.append(cluster.centroid)
                        for idx in cluster.indices:
                            labels[idx] = cluster_id
                        cluster_id += 1
                n_kept = cluster_id

            elif method.lower() == "qbx_and_merge":
                cluster_map = qbx_and_merge(
                    streamlines,
                    thresholds_list,
                    nb_pts=nb_pts,
                    select_randomly=select_randomly,
                )
                n_total = len(cluster_map)
                out_centroids_sl = Streamlines()
                labels = np.full(len(streamlines), -1, dtype=np.int32)
                cluster_id = 0
                for cluster in cluster_map:
                    if len(cluster) >= min_cluster_size:
                        out_centroids_sl.append(cluster.centroid)
                        for idx in cluster.indices:
                            labels[idx] = cluster_id
                        cluster_id += 1
                n_kept = cluster_id

            else:
                cluster_map = QuickBundles(threshold).cluster(streamlines)
                n_total = len(cluster_map)
                centroids_sl = Streamlines(cluster_map.centroids)

                fss = FastStreamlineSearch(streamlines, max_radius=max_radius)
                res = fss.radius_search(centroids_sl, radius=max_radius)

                non_zero_ids, nearest_centroid_id, _ = nearest_from_matrix_col(res)

                raw_labels = np.full(len(streamlines), -1, dtype=np.int32)
                raw_labels[non_zero_ids] = nearest_centroid_id

                out_centroids_sl = Streamlines()
                labels = np.full(len(streamlines), -1, dtype=np.int32)
                new_cluster_id = 0
                for old_cluster_id, centroid in enumerate(cluster_map.centroids):
                    cluster_mask = raw_labels == old_cluster_id
                    if np.sum(cluster_mask) >= min_cluster_size:
                        out_centroids_sl.append(centroid)
                        labels[cluster_mask] = new_cluster_id
                        new_cluster_id += 1
                n_kept = new_cluster_id

            logger.info(
                f"Found {n_total} clusters, kept {n_kept} "
                f"(min_cluster_size={min_cluster_size})"
            )

            sft_out = StatefulTractogram(out_centroids_sl, sft, Space.RASMM)
            save_tractogram(sft_out, centroids_out_path, bbox_valid_check=False)
            np.save(labels_out_path, labels)
            logger.info(f"Centroids saved as {centroids_out_path}")
            logger.info(f"Labels saved as {labels_out_path}")
            logger.info(f"Clustering of {fpath} completed.")

        logger.info("Streamline clustering finished.")



"""
=====================================================================
Reconstruction of the diffusion signal with DTI (single tensor) model
=====================================================================

The diffusion tensor model is a model that describes the diffusion within a
voxel. First proposed by Basser and colleagues :footcite:p:`Basser1994a`, it has
been very influential in demonstrating the utility of diffusion MRI in
characterizing the micro-structure of white matter tissue and of the biophysical
properties of tissue, inferred from local diffusion properties and it is still
very commonly used.

The diffusion tensor models the diffusion signal as:

.. math::

    \\frac{S(\\mathbf{g}, b)}{S_0} = e^{-b\\mathbf{g}^T \\mathbf{D} \\mathbf{g}}

Where $\\mathbf{g}$ is a unit vector in 3 space indicating the direction of
measurement and b are the parameters of measurement, such as the strength and
duration of diffusion-weighting gradient. $S(\\mathbf{g}, b)$ is the
diffusion-weighted signal measured and $S_0$ is the signal conducted in a
measurement with no diffusion weighting. $\\mathbf{D}$ is a positive-definite
quadratic form, which contains six free parameters to be fit. These six
parameters are:

.. math::

    \\mathbf{D} = \\begin{pmatrix} D_{xx} & D_{xy} & D_{xz} \\\\
                       D_{yx} & D_{yy} & D_{yz} \\\\
                       D_{zx} & D_{zy} & D_{zz} \\\\ \\end{pmatrix}

This matrix is a variance/covariance matrix of the diffusivity along the three
spatial dimensions. Note that we can assume that diffusivity has antipodal
symmetry, so elements across the diagonal are equal. For example:
$D_{xy} = D_{yx}$. This is why there are only 6 free parameters to estimate
here.

In the following example we show how to reconstruct your diffusion datasets
using a single tensor model.

First import the necessary modules:

``numpy`` is for numerical computation

"""

import numpy as np

"""
``dipy.io.image`` is for loading / saving imaging datasets
``dipy.io.gradients`` is for loading / saving our bvals and bvecs
"""

from dipy.core.gradients import gradient_table
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, save_nifti

"""
``dipy.reconst`` is for the reconstruction algorithms which we use to create
voxel models from the raw data.
"""

import dipy.reconst.dti as dti

"""
``dipy.data`` is used for small datasets that we use in tests and examples.
"""

from dipy.data import get_fnames

"""
``get_fnames`` will download the raw dMRI dataset of a single subject.
The size of the dataset is 87 MBytes. You only need to fetch once. It
will return the file names of our data.
"""

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")

"""
Next, we read the saved dataset. gtab contains a ``GradientTable``
object (information about the gradients e.g. b-values and b-vectors).
"""

data, affine = load_nifti(hardi_fname)

bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

print(f"data.shape {data.shape}")

"""
data.shape ``(81, 106, 76, 160)``

First of all, we mask and crop the data. This is a quick way to avoid
calculating Tensors on the background of the image. This is done using DIPY_'s
``mask`` module.
"""
from dipy.segment.mask import bounding_box, crop, median_otsu

maskdata, mask = median_otsu(
    data, vol_idx=range(10, 50), median_radius=3, numpass=1, dilate=2
)

mins, maxs = bounding_box(mask)

"""
The ``bounding_box`` function returns the minimum and maximum indices of the
non-zero voxels in every dimension. We use these indices to crop the data
and mask to the smallest possible region that contains all the non-zero voxels.
"""

maskdata = crop(maskdata, mins, maxs)
mask = crop(mask, mins, maxs)

print(f"maskdata.shape {maskdata.shape}")

"""
maskdata.shape ``(72, 87, 59, 160)``

Now that we have prepared the datasets we can go forward with the voxel
reconstruction. First, we instantiate the Tensor model in the following way.
"""

tenmodel = dti.TensorModel(gtab, fit_method="WLS")

"""
The ``fit_method`` argument gives the method that will be used when fitting the
data. Several options are available, such as weighted least squares ``WLS``
(default), non-linear least squares ``NLLS``, as well as robust fitting methods
such as ``RWLS`` and ``RNLLS`` as in :footcite:t:`Coveney2025`.
"""

"""
Fitting the data is very simple. We just need to call the fit method of the
TensorModel in the following way:
"""

tenfit = tenmodel.fit(maskdata)

"""
The fit method creates a ``TensorFit`` object which contains the fitting
parameters and other attributes of the model. You can recover the 6 values
of the triangular matrix representing the tensor D. By default, in DIPY, values
are ordered as (Dxx, Dxy, Dyy, Dxz, Dyz, Dzz). The ``tensor_vals`` variable
defined below is a 4D data with last dimension of size 6.
"""

tensor_vals = dti.lower_triangular(tenfit.quadratic_form)

r"""
You can also recover other metrics from the model. For example we can generate
fractional anisotropy (FA) from the eigen-values of the tensor. FA is used to
characterize the degree to which the distribution of diffusion in a voxel is
directional. That is, whether there is relatively unrestricted diffusion in one
particular direction.

Mathematically, FA is defined as the normalized variance of the eigen-values of
the tensor:

.. math::

        FA = \sqrt{\frac{1}{2} \cdot \frac{(\lambda_1-\lambda_2)^2 +
            (\lambda_1-\lambda_3)^2 + (\lambda_2-\lambda_3)^2}
            {\lambda_1^2 + \lambda_2^2 + \lambda_3^2}}

Where $\lambda_1$, $\lambda_2$ and $\lambda_3$ are the eigen-values of the
tensor.

Note that FA should be interpreted carefully. It may be an indication of
the density of packing of fibers in a voxel, and the amount of myelin wrapping
these axons, but it is not always a measure of "tissue integrity". For example,
FA may decrease in locations in which there is fanning of white matter fibers,
or where more than one population of white matter fibers crosses.
"""

print("Computing anisotropy measures (FA, MD, RGB)")
from dipy.reconst.dti import color_fa, fractional_anisotropy

FA = fractional_anisotropy(tenfit.evals)

"""
In the background of the image the fitting will not be accurate there is no
signal and possibly we will find FA values with nans (not a number). We can
easily remove these in the following way.
"""

FA[np.isnan(FA)] = 0

"""
Saving the FA images is very easy using nibabel_. We need the FA volume and the
affine matrix which transform the image's coordinates to the world coordinates.
Here, we choose to save the FA in ``float32``.
"""

save_nifti("tensor_fa.nii.gz", FA.astype(np.float32), affine)

"""
You can now see the result with any nifti viewer or check it slice by slice
using matplotlib_'s ``imshow``. In the same way you can save the eigen values,
the eigen vectors or any other properties of the tensor.
"""

save_nifti("tensor_evecs.nii.gz", tenfit.evecs.astype(np.float32), affine)

"""
Other tensor statistics can be calculated from the ``tenfit`` object. For
example, a commonly calculated statistic is the mean diffusivity (MD). This is
simply the mean of the  eigenvalues of the tensor. Since FA is a normalized
measure of variance and MD is the mean, they are often used as complimentary
measures. In DIPY, there are two equivalent ways to calculate the mean
diffusivity. One is by calling the ``mean_diffusivity`` module function on the
eigen-values of the ``TensorFit`` class instance:
"""

MD1 = dti.mean_diffusivity(tenfit.evals)
save_nifti("tensors_md.nii.gz", MD1.astype(np.float32), affine)

"""
The other is to call the ``TensorFit`` class method:
"""

MD2 = tenfit.md

"""
Obviously, the quantities are identical.

We can also compute the colored FA or RGB-map :footcite:p:`Pajevic1999`. First,
we make sure that the FA is scaled between 0 and 1, we compute the RGB map and
save it.
"""

FA = np.clip(FA, 0, 1)
RGB = color_fa(FA, tenfit.evecs)
save_nifti("tensor_rgb.nii.gz", np.array(255 * RGB, "uint8"), affine)

"""
Let's try to visualize the tensor ellipsoids of a small rectangular
area in an axial slice of the splenium of the corpus callosum (CC).
"""

print("Computing tensor ellipsoids in a part of the splenium of the CC")

from dipy.data import get_sphere

sphere = get_sphere(name="repulsion724")

from dipy.viz import actor, window

# Enables/disables interactive visualization
interactive = False

scene = window.Scene()

evals = tenfit.evals[13:43, 44:74, 28:29]
evecs = tenfit.evecs[13:43, 44:74, 28:29]

"""
We can color the ellipsoids using the ``color_fa`` values that we calculated
above. In this example we additionally normalize the values to increase the
contrast.
"""

cfa = RGB[13:43, 44:74, 28:29]
cfa /= cfa.max()

scene.add(
    actor.tensor_slicer(evals, evecs, scalar_colors=cfa, sphere=sphere, scale=0.3)
)

print("Saving illustration as tensor_ellipsoids.png")
window.record(
    scene=scene, n_frames=1, out_path="tensor_ellipsoids.png", size=(600, 600)
)
if interactive:
    window.show(scene)

"""
.. rst-class:: centered small fst-italic fw-semibold

Tensor Ellipsoids.
"""

scene.clear()

"""
Finally, we can visualize the tensor Orientation Distribution Functions
for the same area as we did with the ellipsoids.
"""

tensor_odfs = tenmodel.fit(data[20:50, 55:85, 38:39]).odf(sphere)

odf_actor = actor.odf_slicer(tensor_odfs, sphere=sphere, scale=0.5, colormap=None)
scene.add(odf_actor)
print("Saving illustration as tensor_odfs.png")
window.record(scene=scene, n_frames=1, out_path="tensor_odfs.png", size=(600, 600))
if interactive:
    window.show(scene)

"""
.. rst-class:: centered small fst-italic fw-semibold

Tensor ODFs.


Note that while the tensor model is an accurate and reliable model of the
diffusion signal in the white matter, it has the drawback that it only has one
principal diffusion direction. Therefore, in locations in the brain that
contain multiple fiber populations crossing each other, the tensor model may
indicate that the principal diffusion direction is intermediate to these
directions. Therefore, using the principal diffusion direction for tracking in
these locations may be misleading and may lead to errors in defining the
tracks. Fortunately, other reconstruction methods can be used to represent the
diffusion and fiber orientations in those locations. These are presented in
other examples.

References
----------
.. footbibliography::

"""



import nibabel as nib
import numpy as np
from numpy.testing import assert_, assert_equal

import dipy.data as dpd
from dipy.denoise.adaptive_soft_matching import adaptive_soft_matching
from dipy.denoise.nlmeans import nlmeans
from dipy.denoise.noise_estimate import estimate_sigma
from dipy.testing.decorators import set_random_number_generator


def test_ascm_static():
    S0 = 100 * np.ones((20, 20, 20), dtype="f8")
    S0n1 = nlmeans(S0, sigma=0, rician=False, patch_radius=1, block_radius=1)
    S0n2 = nlmeans(S0, sigma=0, rician=False, patch_radius=2, block_radius=1)
    S0n = adaptive_soft_matching(S0, S0n1, S0n2, 0)

    assert_equal(np.round(S0n.mean()), 100)
    assert_(np.abs(S0n.mean() - S0.mean()) < 1.0)

    close_values = np.abs(S0n - S0) < 10.0
    assert_(np.sum(close_values) / S0.size > 0.95)


@set_random_number_generator()
def test_ascm_random_noise(rng):
    S0 = 100 + 2 * rng.standard_normal((22, 23, 30))
    S0n1 = nlmeans(S0, sigma=1, rician=False, patch_radius=1, block_radius=1)
    S0n2 = nlmeans(S0, sigma=1, rician=False, patch_radius=2, block_radius=1)
    S0n = adaptive_soft_matching(S0, S0n1, S0n2, 1)

    assert_(np.abs(S0n.mean() - S0.mean()) < 3.0)
    assert_(np.isfinite(S0n).all())
    assert_(not np.isnan(S0n).any())
    assert_(S0n.std() < S0.std() * 1.5)
    assert_(S0n.min() > 0)
    assert_(S0n.max() < 200)

    reasonable_values = np.abs(S0n - S0n.mean()) < 3 * S0n.std()
    assert_(np.mean(reasonable_values) > 0.95)


@set_random_number_generator()
def test_ascm_rmse_with_nlmeans(rng):
    # checks the smoothness
    S0 = np.ones((30, 30, 30)) * 100
    S0[10:20, 10:20, 10:20] = 50
    S0[20:30, 20:30, 20:30] = 0
    S0_noise = S0 + 20 * rng.standard_normal((30, 30, 30))
    print("Original RMSE", np.sum(np.abs(S0 - S0_noise)) / np.sum(S0))

    S0n1 = nlmeans(S0_noise, sigma=400, rician=False, patch_radius=1, block_radius=1)
    print("Smaller patch RMSE", np.sum(np.abs(S0 - S0n1)) / np.sum(S0))
    S0n2 = nlmeans(S0_noise, sigma=400, rician=False, patch_radius=2, block_radius=2)
    print("Larger patch RMSE", np.sum(np.abs(S0 - S0n2)) / np.sum(S0))
    S0n = adaptive_soft_matching(S0, S0n1, S0n2, 400)
    print("ASCM RMSE", np.sum(np.abs(S0 - S0n)) / np.sum(S0))

    assert_(
        np.sum(np.abs(S0 - S0n)) / np.sum(S0) < np.sum(np.abs(S0 - S0n1)) / np.sum(S0)
    )
    assert_(
        np.sum(np.abs(S0 - S0n)) / np.sum(S0)
        < np.sum(np.abs(S0 - S0_noise)) / np.sum(S0)
    )
    assert_(90 < np.mean(S0n) < 110)


@set_random_number_generator()
def test_sharpness(rng):
    # check the edge-preserving nature
    S0 = np.ones((30, 30, 30)) * 100
    S0[10:20, 10:20, 10:20] = 50
    S0[20:30, 20:30, 20:30] = 0
    S0_noise = S0 + 20 * rng.standard_normal((30, 30, 30))
    S0n1 = nlmeans(S0_noise, sigma=400, rician=False, patch_radius=1, block_radius=1)
    edg1 = np.abs(np.mean(S0n1[8, 10:20, 10:20] - S0n1[12, 10:20, 10:20]) - 50)
    print("Edge gradient smaller patch", edg1)
    S0n2 = nlmeans(S0_noise, sigma=400, rician=False, patch_radius=2, block_radius=2)
    edg2 = np.abs(np.mean(S0n2[8, 10:20, 10:20] - S0n2[12, 10:20, 10:20]) - 50)
    print("Edge gradient larger patch", edg2)
    S0n = adaptive_soft_matching(S0, S0n1, S0n2, 400)
    edg = np.abs(np.mean(S0n[8, 10:20, 10:20] - S0n[12, 10:20, 10:20]) - 50)
    print("Edge gradient ASCM", edg)

    assert_(edg2 > edg1)
    assert_(edg2 > edg)
    assert_(np.abs(edg1 - edg) < 1.5)


def test_ascm_accuracy():
    f_name = dpd.get_fnames(name="ascm_test")
    test_ascm_data_ref = np.asanyarray(nib.load(f_name).dataobj)
    test_data = np.asanyarray(nib.load(dpd.get_fnames(name="aniso_vox")).dataobj)

    # the test data was constructed in this manner
    mask = test_data > 50
    sigma = estimate_sigma(test_data, N=4).item()

    den_small = nlmeans(
        test_data, sigma=sigma, mask=mask, patch_radius=1, block_radius=1, rician=True
    )

    den_large = nlmeans(
        test_data, sigma=sigma, mask=mask, patch_radius=2, block_radius=1, rician=True
    )

    S0n = np.array(adaptive_soft_matching(test_data, den_small, den_large, sigma))

    S0n_masked = S0n[mask]
    ref_masked = test_ascm_data_ref[mask]
    orig_masked = test_data[mask]

    correlation = np.corrcoef(S0n_masked.flatten(), ref_masked.flatten())[0, 1]
    assert_(correlation > 0.9)

    mean_diff = np.abs(np.mean(S0n_masked) - np.mean(ref_masked))
    assert_(mean_diff < 10.0)

    assert_(np.var(S0n_masked) < np.var(orig_masked))



from copy import deepcopy
import gzip
import logging
from pathlib import Path
import time

import nibabel as nib
import numpy as np

from dipy.io.stateful_surface import StatefulSurface
from dipy.io.utils import (
    Origin,
    Space,
    get_reference_info,
    split_filename_extension,
)
from dipy.io.vtk import (
    get_polydata_triangles,
    get_polydata_vertices,
    load_polydata,
    save_polydata,
)
from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

vtk, have_vtk, setup_module = optional_package(
    "vtk", min_version="9.0.0", max_version="9.1.0"
)

if have_vtk:
    import vtk
    import vtk.util.numpy_support as ns


def load_surface(
    fname,
    reference,
    *,
    to_space=Space.RASMM,
    to_origin=Origin.NIFTI,
    bbox_valid_check=True,
    from_space=None,
    from_origin=None,
    gifti_in_freesurfer=False,
):
    """Load the stateful surface from any format (vtk/vtp/obj/stl/ply/gii/pial)

    Parameters
    ----------
    filename : string or Path
        Filename with valid extension
    reference : Nifti or Trk filename, Nifti1Image or TrkFile, Nifti1Header or
        trk.header (dict), or 'same' if the input is a trk file.
        Reference that provides the spatial attribute.
        Typically a nifti-related object from the native diffusion used for
        surface generation
    to_space : Enum (dipy.io.utils.Space), optional
        Space to which the surface will be transformed after loading
    to_origin : Enum (dipy.io.utils.Origin), optional
        Origin to which the surface will be transformed after loading
            NIFTI standard, default (center of the voxel)
            TRACKVIS standard (corner of the voxel)
    bbox_valid_check : bool, optional
        Verification for negative voxel coordinates or values above the
        volume dimensions. Default is True, to enforce valid file.
    from_space : Enum (dipy.io.utils.Space), optional
        Space to which the surface was transformed before saving.
        Help for software compatibility. If None, assumes RASMM.
    from_origin : Enum (dipy.io.utils.Origin), optional
        Origin to which the surface was transformed before saving.
        Help for software compatibility. If None, assumes NIFTI.
    gifti_in_freesurfer : bool, optional
        Whether the gifti file is in freesurfer reference space.

    Raises
    ------
    ValueError
        If the bounding box is not valid in voxel space.

    Returns
    -------
    output : StatefulSurface
        The surface to load (must have been saved properly)
    """
    vtk_ext = [".vtk", ".vtp", ".obj", ".stl", ".ply"]
    freesurfer_ext = [".gii", ".gii.gz", ".pial", ".nofix", ".orig", ".smoothwm", ".T1"]

    name, ext = split_filename_extension(fname)
    if ext.lower() not in freesurfer_ext + vtk_ext:
        logging.error("Input extension is not one of the supported format.")
        return False

    if to_space not in Space:
        logging.error(
            f"Space MUST be one of the {len(Space)} choices:"
            f" {list(Space.__members__.keys())}."
        )
        return False

    if to_origin not in Origin:
        logging.error(
            f"Origin MUST be one of the {len(Origin)} choices:"
            f" {list(Origin.__members__.keys())}."
        )
        return False

    if reference == "same":
        if ext in [".gii", ".gii.gz"]:
            reference = fname
        else:
            logging.error(
                'Reference must be provided, "same" is only available for GII file.'
            )
            return False

    timer = time.time()
    metadata = None
    data_per_vertex = None
    if ext in [".gii", ".gii.gz"]:
        data = load_gifti(fname)
        if gifti_in_freesurfer:
            data = (apply_freesurfer_transform(data[0], reference, inv=True), data[1])
        vertices = np.array(data[0])
        faces = np.array(data[1])
    elif ext in [".vtk", ".vtp", ".obj", ".stl", ".ply"]:
        data = load_polydata(fname)
        vertices = get_polydata_vertices(data)
        faces = get_polydata_triangles(data)
        vertex_data = data.GetPointData()
        scalar_names = [
            vertex_data.GetArrayName(i) for i in range(vertex_data.GetNumberOfArrays())
        ]
        if scalar_names:
            for name in scalar_names:
                scalar = data.GetPointData().GetScalars(name)
                if name in data_per_vertex:
                    logging.warning(
                        f"Scalar {name} already in data_per_vertex, overwriting"
                    )
                data_per_vertex[name] = ns.vtk_to_numpy(scalar)
    else:
        data = load_pial(fname, return_meta=True)
        data, metadata = data[0:2], data[2]

        data = (apply_freesurfer_transform(data[0], reference, inv=True), data[1])
        if from_space is not None or from_origin is not None:
            logging.warning(
                "from_space and from_origin are ignored when loading pial files."
            )
        from_space = Space.RASMM
        from_origin = Origin.NIFTI
        vertices = np.array(data[0])
        faces = np.array(data[1])

    from_space = Space.RASMM if from_space is None else from_space
    from_origin = Origin.NIFTI if from_origin is None else from_origin
    sfs = StatefulSurface(
        vertices,
        faces,
        reference,
        space=from_space,
        origin=from_origin,
        data_per_vertex=data_per_vertex,
    )
    if isinstance(metadata, dict):
        sfs.fs_metadata = metadata
    elif isinstance(metadata, nib.filebasedimages.FileBasedHeader):
        sfs.gii_header = metadata

    logging.debug(
        f"Load {fname} with {len(sfs)} vertices in {round(time.time() - timer, 3)}"
        f" seconds."
    )

    if bbox_valid_check and not sfs.is_bbox_in_vox_valid():
        raise ValueError(
            "Bounding box is not valid in voxel space, cannot "
            "load a valid file if some coordinates are invalid.\n"
            "Please set bbox_valid_check to False and be careful if processing "
            "the surface/vertices further."
        )

    sfs.to_space(to_space)
    sfs.to_origin(to_origin)

    return sfs


def save_surface(
    sfs,
    fname,
    *,
    to_space=Space.RASMM,
    to_origin=Origin.NIFTI,
    legacy_vtk_format=False,
    bbox_valid_check=True,
    ref_pial=None,
    ref_gii=None,
    gifti_in_freesurfer=False,
):
    """
    Save the stateful surface to any format (vtk/vtp/obj/stl/ply/gii/pial)

    Parameters
    ----------
    sfs : StatefulSurface
        The surface to save (must have been loaded properly)
    fname : str or Path
        Absolute path of the file.
    to_space : Enum (dipy.io.stateful_surface.Space), optional
        Space to which the surface will be transformed before saving
    to_origin : Enum (dipy.io.stateful_surface.Origin), optional
        Origin to which the surface will be transformed before saving
            NIFTI standard, default (center of the voxel)
            TRACKVIS standard (corner of the voxel)
    legacy_vtk_format : bool, optional
        Whether to save the file in legacy VTK format or not.
    bbox_valid_check : bool, optional
        Verification for negative voxel coordinates or values above the
        volume dimensions. Default is True, to enforce valid file.
    ref_pial : str or Path, optional
        Reference pial file to save the surface in pial format.
        If not provided, the metadata of the input surface is used (if available).
    ref_gii : str or Path, optional
        Reference gii file to save the surface in gii format.
        If not provided, the header of the input surface is used (if available).
    gifti_in_freesurfer : bool, optional
        Whether the gifti file must be saved in freesurfer reference space.

    Raises
    ------
    ValueError
        If the bounding box is not valid in voxel space.
    """
    old_space = deepcopy(sfs.space)
    old_origin = deepcopy(sfs.origin)

    if bbox_valid_check and not sfs.is_bbox_in_vox_valid():
        raise ValueError(
            "Bounding box is not valid in voxel space, cannot "
            "save a valid file if some coordinates are invalid.\n"
            "Please set bbox_valid_check to False and verify the "
            "bounding box of the surface."
        )

    _, ext = split_filename_extension(fname)
    ext = ext.lower()

    if ext in [".vtk", ".vtp", ".obj", ".stl", ".ply"]:
        sfs.to_space(to_space)
        sfs.to_origin(to_origin)

        if sfs.data_per_vertex is not None:
            # Check if rgb, colors, colors, etc. are available
            color_array_name = None
            for key in ["rgb", "colors", "color"]:
                if key in sfs.data_per_vertex:
                    color_array_name = key
                    break
                if key.upper() in sfs.data_per_vertex:
                    color_array_name = key.upper()
                    break

        polydata = sfs.get_polydata()
        if color_array_name is not None:
            color_array = sfs.data_per_vertex[color_array_name]
            if len(color_array) != polydata.GetNumberOfPoints():
                raise ValueError("Array length does not match number of vertices.")
            vtk_array = ns.numpy_to_vtk(
                np.array(color_array), deep=True, array_type=vtk.VTK_UNSIGNED_CHAR
            )
            vtk_array.SetName("RGB")

            if "normal" in color_array_name.lower():
                polydata.GetPointData().SetNormals(vtk_array)
            else:
                try:
                    polydata.GetPointData().SetScalars(vtk_array)
                except ValueError:
                    polydata.GetPointData().AddArray(vtk_array)

        save_polydata(
            polydata, fname, legacy_vtk_format=legacy_vtk_format, color_array_name="RGB"
        )
    elif ext in [".gii", ".gii.gz"]:
        if not hasattr(sfs, "gii_header") and ref_gii is None:
            raise ValueError(
                "Metadata is required to save a gii file.\n"
                "Please provide the reference gii file."
            )

        if ref_gii is not None:
            _, ext = split_filename_extension(ref_gii)
            ext = ext.lower()
            if ext not in [".gii", ".gii.gz"]:
                raise ValueError("Reference gii file must have .gii extension.")
            _, metadata = load_gifti(ref_gii, return_header=True)[-1]
        else:
            metadata = sfs.gii_header

        sfs.to_space(to_space)
        sfs.to_origin(to_origin)
        if gifti_in_freesurfer:
            sfs.vertices = apply_freesurfer_transform(sfs.vertices, sfs, inv=False)
        save_gifti(fname, sfs.vertices, sfs.faces, header=metadata)

    elif ext == ".pial":
        if not hasattr(sfs, "fs_metadata") and ref_pial is None:
            raise ValueError(
                "Metadata is required to save a pial file.\n"
                "Please provide the reference pial file."
            )

        if ref_pial is not None:
            ext = Path(ref_pial).suffix
            if ext != ".pial":
                raise ValueError("Reference pial file must have .pial extension.")
            metadata = load_pial(ref_pial, return_meta=True)[-1]
        else:
            metadata = sfs.fs_metadata

        if to_space is not None or to_origin is not None:
            logging.warning(
                "to_space and to_origin are ignored when loading pial files."
            )
        sfs.to_space(Space.RASMM)
        sfs.to_origin(Origin.NIFTI)

        vertices = apply_freesurfer_transform(sfs.vertices, sfs, inv=False)

        save_pial(fname, vertices, sfs.faces, metadata=metadata)
    else:
        logging.error("Output extension is not one of the supported format.")

    sfs.to_space(old_space)
    sfs.to_origin(old_origin)


@warning_for_keywords()
def load_pial(fname, *, return_meta=False):
    """Load pial file.

    Parameters
    ----------
    fname : str or Path
        Absolute path of the file.
    return_meta : bool, optional
        Whether to read the metadata of the file or not, by default False.

    Returns
    -------
    tuple
        (vertices, faces) if return_meta=False. Otherwise, (vertices, faces,
        metadata).
    """
    try:
        data = nib.freesurfer.read_geometry(fname, read_metadata=return_meta)
    except ValueError:
        try:
            data = nib.freesurfer.read_geometry(fname, read_metadata=False)
            logging.warning("No metadata found, please use a pial file with metadata.")
        except ValueError:
            raise ValueError(f"{fname} provided does not have geometry data.") from None

    return data


def save_pial(fname, vertices, faces, *, metadata=None):
    """Save pial file.

    Parameters
    ----------
    fname : str or Path
        Absolute path of the file.
    vertices : ndarray
        Vertices.
    faces : ndarray
        Faces.
    metadata : dict, optional
        Key-value pairs to encode at the end of the file.
    """
    nib.freesurfer.write_geometry(fname, vertices, faces, volume_info=metadata)


def load_gifti(fname, *, return_header=False):
    """Load gifti file.

    Parameters
    ----------
    fname : str or Path
        Absolute path of the file.

    return_header : bool, optional
        Whether to read the header of the file or not, by default False.
        If True, returns a tuple with vertices, faces and header.

    Returns
    -------
    tuple
        (vertices, faces)
    """

    def reader(fname):
        _, ext = split_filename_extension(fname)
        ext = ext.lower()
        if ext == ".gii.gz":
            with gzip.GzipFile(fname) as gz:
                img = nib.GiftiImage.from_bytes(gz.read())
        else:
            img = nib.load(fname)
        return img

    if return_header:
        gifti_img = reader(fname)
        return gifti_img.agg_data(("pointset", "triangle")), gifti_img.header
    else:
        return reader(fname).agg_data(("pointset", "triangle"))


def save_gifti(fname, vertices, faces, *, header=None):
    """Save gifti file.
    https://netneurolab.github.io/neuromaps/_modules/neuromaps/images.html

    Parameters
    ----------
    fname : str or Path
        Absolute path of the file.
    vertices : ndarray
        Vertices.
    faces : ndarray
        Faces.
    header : nib.filebasedimages.FileBasedHeader
        Valid header for the gifti file, typically loaded from a reference GII
    """
    vert = nib.gifti.GiftiDataArray(
        vertices,
        "NIFTI_INTENT_POINTSET",
        "NIFTI_TYPE_FLOAT32",
        coordsys=nib.gifti.GiftiCoordSystem(3, 3),
    )
    tri = nib.gifti.GiftiDataArray(faces, "NIFTI_INTENT_TRIANGLE", "NIFTI_TYPE_INT32")
    img = nib.GiftiImage(darrays=[vert, tri])
    nib.save(img, fname)


def apply_freesurfer_transform(vertices, reference, *, inv=False):
    """
    Apply the freesurfer transform to the vertices to bring them to RASMM space
    or to bring them back to the original space.

    Parameters
    ----------
    vertices : ndarray
        Vertices to transform.
    reference : str or Path
        Reference file to get the transform from.
    inv : bool, optional
        True if loading the surface, False if saving the surface.
    """

    affine, dimensions, _, _ = get_reference_info(reference)
    center_volume = np.array(dimensions) / 2
    vertices_copy = vertices.copy()
    if inv:
        xform_translation = np.dot(affine[0:3, 0:3], center_volume) + affine[0:3, 3]
    else:
        xform_translation = -(np.dot(affine[0:3, 0:3], center_volume) + affine[0:3, 3])

    return vertices_copy + xform_translation



import json
import os
from pathlib import Path
import sys
import warnings

import numpy as np

from dipy.core.gradients import unique_bvals_tolerance
from dipy.direction.peaks import PeaksAndMetrics
from dipy.reconst._force_search import search_inner_product as _cython_search
from dipy.reconst.base import ReconstFit, ReconstModel
from dipy.reconst.multi_voxel import multi_voxel_fit
from dipy.reconst.shm import sf_to_sh
from dipy.sims.force import (
    generate_force_simulations,
    get_default_diffusivity_config,
    load_force_simulations,
    save_force_simulations,
)
from dipy.utils.logging import logger

# Named constants
EPSILON = 1e-12


def _get_force_cache_dir():
    """Return the FORCE simulation cache directory inside .dipy.

    Uses ``DIPY_HOME`` environment variable if set, otherwise defaults
    to ``~/.dipy/force_simulations``.

    Returns
    -------
    cache_dir : Path
        Path to the cache directory (created if it does not exist).
    """
    if "DIPY_HOME" in os.environ:
        dipy_home = Path(os.environ["DIPY_HOME"])
    else:
        dipy_home = Path("~").expanduser() / ".dipy"
    cache_dir = dipy_home / "force_simulations"
    cache_dir.mkdir(parents=True, exist_ok=True)
    return cache_dir


def _gtab_matches(entry_bvals, entry_bvecs, gtab, *, bval_tol=10.0, bvec_tol=1e-3):
    """Check whether stored bvals/bvecs match a GradientTable.

    Parameters
    ----------
    entry_bvals : list
        Stored b-values from the cache registry.
    entry_bvecs : list of list
        Stored b-vectors from the cache registry.
    gtab : GradientTable
        Gradient table to compare against.
    bval_tol : float, optional
        Absolute tolerance for b-value comparison.
    bvec_tol : float, optional
        Absolute tolerance for b-vector coordinate comparison.

    Returns
    -------
    match : bool
        True if the stored and passed bvals/bvecs agree within tolerance.
    """
    stored_bvals = np.asarray(entry_bvals, dtype=np.float64)
    stored_bvecs = np.asarray(entry_bvecs, dtype=np.float64)
    current_bvals = np.asarray(gtab.bvals, dtype=np.float64)
    current_bvecs = np.asarray(gtab.bvecs, dtype=np.float64)

    if stored_bvals.shape != current_bvals.shape:
        return False
    if stored_bvecs.shape != current_bvecs.shape:
        return False

    return np.allclose(stored_bvals, current_bvals, atol=bval_tol) and np.allclose(
        stored_bvecs, current_bvecs, atol=bvec_tol
    )


def _diffusivity_matches(entry_config, current_config):
    """Check whether two diffusivity configurations are equivalent.

    Parameters
    ----------
    entry_config : dict
        Stored diffusivity configuration.
    current_config : dict
        Current diffusivity configuration.

    Returns
    -------
    match : bool
        True if all keys and values are identical.
    """
    if set(entry_config.keys()) != set(current_config.keys()):
        return False
    for key in entry_config:
        stored = entry_config[key]
        current = current_config[key]
        # Both may be lists/tuples (ranges) or scalars
        if isinstance(stored, (list, tuple)):
            if not isinstance(current, (list, tuple)):
                return False
            if len(stored) != len(current):
                return False
            if not all(np.isclose(s, c) for s, c in zip(stored, current)):
                return False
        else:
            if not np.isclose(stored, current):
                return False
    return True


def _load_cache_registry(cache_dir):
    """Load the cache registry JSON from *cache_dir*.

    Returns an empty list if the file does not exist yet.
    """
    registry_path = cache_dir / "cache_registry.json"
    if registry_path.exists():
        with open(registry_path, "r") as f:
            return json.load(f)
    return []


def _save_cache_registry(cache_dir, registry):
    """Persist *registry* as JSON in *cache_dir*."""
    registry_path = cache_dir / "cache_registry.json"
    with open(registry_path, "w") as f:
        json.dump(registry, f, indent=2)


def _locked_registry_update(cache_dir, update_fn):
    """Read-modify-write the cache registry under an exclusive file lock.

    Parameters
    ----------
    cache_dir : Path
        Cache directory.
    update_fn : callable
        Function that receives the current registry list and returns the
        updated list.
    """
    lock_path = cache_dir / "cache_registry.lock"
    with open(lock_path, "w") as lock_fh:
        if sys.platform == "win32":
            import msvcrt

            msvcrt.locking(lock_fh.fileno(), msvcrt.LK_LOCK, 1)
            try:
                registry = _load_cache_registry(cache_dir)
                registry = update_fn(registry)
                _save_cache_registry(cache_dir, registry)
            finally:
                lock_fh.seek(0)
                msvcrt.locking(lock_fh.fileno(), msvcrt.LK_UNLCK, 1)
        else:
            import fcntl

            fcntl.flock(lock_fh, fcntl.LOCK_EX)
            try:
                registry = _load_cache_registry(cache_dir)
                registry = update_fn(registry)
                _save_cache_registry(cache_dir, registry)
            finally:
                fcntl.flock(lock_fh, fcntl.LOCK_UN)


def _find_cached_simulation(cache_dir, gtab, diffusivity_config, num_simulations):
    """Search the registry for a simulation matching the given parameters.

    Parameters
    ----------
    cache_dir : Path
        Cache directory.
    gtab : GradientTable
        Gradient table.
    diffusivity_config : dict
        Diffusivity ranges used for generation.
    num_simulations : int
        Number of simulations requested.

    Returns
    -------
    path : str or None
        Path to the cached ``.npz`` file, or None if no match found.
    """
    registry = _load_cache_registry(cache_dir)
    for entry in registry:
        if entry["num_simulations"] != num_simulations:
            continue
        if not _gtab_matches(entry["bvals"], entry["bvecs"], gtab):
            continue
        if not _diffusivity_matches(entry["diffusivity_config"], diffusivity_config):
            continue
        candidate = cache_dir / entry["filename"]
        if candidate.exists():
            return str(candidate)
    return None


def _register_cached_simulation(
    cache_dir, gtab, diffusivity_config, num_simulations, filename
):
    """Add a new entry to the cache registry.

    Parameters
    ----------
    cache_dir : Path
        Cache directory.
    gtab : GradientTable
        Gradient table.
    diffusivity_config : dict
        Diffusivity ranges.
    num_simulations : int
        Number of simulations.
    filename : str
        Filename of the saved ``.npz`` inside *cache_dir*.
    """

    # Convert numpy types to plain Python for JSON serialisation
    def _to_json(obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, (np.floating, np.integer)):
            return obj.item()
        if isinstance(obj, tuple):
            return list(obj)
        return obj

    config_json = {}
    for k, v in diffusivity_config.items():
        config_json[k] = _to_json(v)

    entry = {
        "bvals": np.asarray(gtab.bvals, dtype=np.float64).tolist(),
        "bvecs": np.asarray(gtab.bvecs, dtype=np.float64).tolist(),
        "diffusivity_config": config_json,
        "num_simulations": int(num_simulations),
        "filename": filename,
    }

    def _append(registry):
        registry.append(entry)
        return registry

    _locked_registry_update(cache_dir, _append)


class SignalIndex:
    """Index for inner product similarity search.

    Uses optimized Cython BLAS for fast matrix multiplication
    and streaming heap for memory-efficient top-k selection.

    Parameters
    ----------
    d : int
        Dimension of vectors.
    """

    def __init__(self, d):
        if d <= 0:
            raise ValueError(f"Dimension must be positive, got {d}")
        self.d = int(d)
        self.ntotal = 0
        self._xb = None

    def add(self, x):
        """Add vectors to the index.

        Parameters
        ----------
        x : array-like (n, d)
            Vectors to add, will be converted to float32 C-contiguous.

        Notes
        -----
        Each call reallocates the internal array via ``np.vstack``.  This
        method is designed for a single bulk load; repeated small ``add``
        calls will exhibit O(n²) memory allocation cost.
        """
        x = np.ascontiguousarray(x, dtype=np.float32)

        if x.ndim == 1:
            if len(x) != self.d:
                raise ValueError(
                    f"Vector dimension {len(x)} != index dimension {self.d}"
                )
            x = x.reshape(1, -1)

        if x.ndim != 2:
            raise ValueError(f"Expected 1D or 2D array, got {x.ndim}D")

        if x.shape[1] != self.d:
            raise ValueError(
                f"Vector dimension {x.shape[1]} != index dimension {self.d}"
            )

        if self._xb is None:
            self._xb = x.copy()
        else:
            self._xb = np.vstack([self._xb, x])

        self.ntotal = len(self._xb)

    def search(self, x, k):
        """Search for k nearest neighbors by inner product.

        Parameters
        ----------
        x : array-like (n, d) or (d,)
            Query vectors.
        k : int
            Number of neighbors.

        Returns
        -------
        distances : ndarray (n, k)
            Inner products (descending order).
        indices : ndarray (n, k)
            Neighbor indices.
        """
        if self.ntotal == 0:
            raise RuntimeError("Cannot search empty index")

        x = np.ascontiguousarray(x, dtype=np.float32)

        if x.ndim == 1:
            if len(x) != self.d:
                raise ValueError(
                    f"Query dimension {len(x)} != index dimension {self.d}"
                )
            x = x.reshape(1, -1)

        if x.ndim != 2:
            raise ValueError(f"Expected 1D or 2D array, got {x.ndim}D")

        if x.shape[1] != self.d:
            raise ValueError(
                f"Query dimension {x.shape[1]} != index dimension {self.d}"
            )

        if k <= 0:
            raise ValueError(f"k must be positive, got {k}")

        if k > self.ntotal:
            warnings.warn(
                f"k={k} exceeds index size ({self.ntotal}); "
                f"clamping to {self.ntotal}.",
                UserWarning,
                stacklevel=2,
            )
            k = self.ntotal

        # Use optimized Cython search (SciPy BLAS + streaming heap)
        distances, indices = _cython_search(x, self._xb, k)

        return distances, indices

    def reset(self):
        """Remove all vectors from the index."""
        self._xb = None
        self.ntotal = 0

    def __repr__(self):
        return f"SignalIndex(d={self.d}, ntotal={self.ntotal})"


def normalize_signals(signals):
    """L2-normalize signal array for cosine similarity search.

    Parameters
    ----------
    signals : ndarray (N, M)
        Signal array with N samples and M measurements.

    Returns
    -------
    normalized : ndarray (N, M)
        L2-normalized signals.
    """
    signals = np.asarray(signals, dtype=np.float32)
    norms = np.linalg.norm(signals, axis=1, keepdims=True)
    norms[norms == 0] = 1.0
    return np.ascontiguousarray(signals / norms)


def create_signal_index(signals_norm):
    """Create index for cosine similarity search.

    Parameters
    ----------
    signals_norm : ndarray (N, M)
        L2-normalized library signals.

    Returns
    -------
    index : SignalIndex
        Search index.
    """
    dimension = signals_norm.shape[1]
    index = SignalIndex(dimension)
    index.add(signals_norm)
    return index


def softmax_stable(x, *, axis=1):
    """Numerically stable softmax.

    Parameters
    ----------
    x : ndarray
        Input array.
    axis : int, optional
        Axis along which to compute softmax.

    Returns
    -------
    softmax : ndarray
        Softmax probabilities.
    """
    x = x - np.max(x, axis=axis, keepdims=True)
    ex = np.exp(x)
    return ex / np.sum(ex, axis=axis, keepdims=True)


def compute_uncertainty_ambiguity(scores):
    """Compute uncertainty and ambiguity metrics from match scores.

    Parameters
    ----------
    scores : ndarray (N, K)
        Penalized scores for K neighbors.

    Returns
    -------
    uncertainty : ndarray (N,)
        IQR of scores.
    ambiguity : ndarray (N,)
        Fraction above half-max.
    """
    p75 = np.percentile(scores, 75, axis=1)
    p25 = np.percentile(scores, 25, axis=1)
    uncertainty = (p75 - p25).astype(np.float32)

    s_max = np.max(scores, axis=1)
    s_min = np.min(scores, axis=1)
    half = 0.5 * (s_max + s_min)
    ambiguity = (np.sum(scores > half[:, None], axis=1) / scores.shape[1]).astype(
        np.float32
    )
    return uncertainty, ambiguity


MICRO_PARAMS = (
    "fa",
    "md",
    "rd",
    "wm_fraction",
    "gm_fraction",
    "csf_fraction",
    "num_fibers",
    "dispersion",
    "nd",
    "ufa_voxel",
    "ak",
    "rk",
    "mk",
    "kfa",
)


def _weighted_percentile(vals, weights, q):
    """Batch weighted percentile.

    Parameters
    ----------
    vals : ndarray (N, K)
        Parameter values across K neighbors per voxel.
    weights : ndarray (N, K)
        Posterior weights.
    q : float
        Quantile in [0, 1].

    Returns
    -------
    result : ndarray (N,)
        Weighted percentile for each voxel.
    """
    idx = np.argsort(vals, axis=1)
    sv = np.take_along_axis(vals, idx, axis=1)
    sw = np.take_along_axis(weights, idx, axis=1)
    cw = np.cumsum(sw, axis=1)
    cw = cw / (cw[:, -1:] + EPSILON)
    j = np.argmax(cw >= q, axis=1)
    return sv[np.arange(sv.shape[0]), j]


def _fwhm_kde_batch(vals, weights, *, n_grid=100, batch_size=2000):
    """Batch FWHM via weighted KDE.

    Estimates the full width at half maximum of the weighted posterior
    density for each voxel using a Gaussian kernel density estimate.

    Parameters
    ----------
    vals : ndarray (N, K)
        Parameter values across K neighbors.
    weights : ndarray (N, K)
        Posterior weights.
    n_grid : int, optional
        Number of grid points for KDE evaluation.
    batch_size : int, optional
        Internal batch size to limit memory usage.

    Returns
    -------
    fwhm : ndarray (N,)
        FWHM in parameter units.
    """
    N, K = vals.shape
    fwhm = np.empty(N, dtype=np.float32)

    for start in range(0, N, batch_size):
        end = min(start + batch_size, N)
        vals_batch = vals[start:end].astype(np.float32)
        weights_batch = weights[start:end].astype(np.float32)
        n = end - start

        lo = vals_batch.min(axis=1)
        hi = vals_batch.max(axis=1)
        span = hi - lo
        pad = 0.15 * (span + EPSILON)
        lo_p, hi_p = lo - pad, hi + pad
        span_p = hi_p - lo_p

        t = np.linspace(0, 1, n_grid, dtype=np.float32)
        grid = lo_p[:, None] + t[None, :] * span_p[:, None]

        bw = span_p / max(K**0.5, 4.0)
        bw = np.maximum(bw, EPSILON)

        diff = (grid[:, :, None] - vals_batch[:, None, :]) / bw[:, None, None]
        density = np.sum(np.exp(-0.5 * diff**2) * weights_batch[:, None, :], axis=2)

        peak = density.max(axis=1)
        half_peak = 0.5 * peak
        above = density >= half_peak[:, None]

        any_above = np.any(above, axis=1)
        first = np.argmax(above, axis=1)
        last = n_grid - 1 - np.argmax(above[:, ::-1], axis=1)

        f = grid[np.arange(n), last] - grid[np.arange(n), first]
        f[~any_above] = 0.0
        fwhm[start:end] = f

    return fwhm


def compute_microstructure_uncertainty_ambiguity(vals, weights, prior_range):
    """Compute per-microstructure uncertainty and ambiguity.

    Uncertainty is the weighted interquartile range of the posterior density,
    normalized by the prior range. Ambiguity is the full width at half
    maximum (FWHM) of the weighted posterior density, similarly normalized.
    Both are in [0, 1] where 0 means no spread and 1 means the posterior
    spans the entire prior range.

    Parameters
    ----------
    vals : ndarray (N, K)
        Parameter values for K neighbors per voxel.
    weights : ndarray (N, K)
        Posterior weights (softmax).
    prior_range : float
        Max minus min of this parameter across the simulation library.

    Returns
    -------
    uncertainty : ndarray (N,)
        Weighted IQR / prior_range (fraction in [0, 1]).
    ambiguity : ndarray (N,)
        FWHM / prior_range (fraction in [0, 1]).
    """
    q75 = _weighted_percentile(vals, weights, 0.75)
    q25 = _weighted_percentile(vals, weights, 0.25)
    uncertainty = ((q75 - q25) / (prior_range + EPSILON)).astype(np.float32)

    fwhm = _fwhm_kde_batch(vals, weights)
    ambiguity = (fwhm / (prior_range + EPSILON)).astype(np.float32)

    return uncertainty, ambiguity


def postprocess_peaks(preds, target_sphere, fracs):
    original_shape = preds.shape[:-1]
    preds_flat = preds.reshape(-1, preds.shape[-1])
    fracs_flat = fracs.reshape(-1, fracs.shape[-1])

    n_voxels = preds_flat.shape[0]
    vertices = target_sphere.vertices

    # Initialize outputs using the total number of voxels
    peaks_output = np.zeros((n_voxels, 5, 3), dtype=np.float32)
    peak_indices = np.full((n_voxels, 5), -1, dtype=np.int32)
    peak_vals = np.zeros((n_voxels, 5), dtype=np.float32)

    for i in range(n_voxels):
        mask = preds_flat[i] == 1
        coords = vertices[mask]
        indices = np.where(mask)[0]

        num = min(len(coords), 5)
        if num > 0:
            peaks_output[i, :num] = coords[:num]
            peak_indices[i, :num] = indices[:num]

        num_fracs = min(5, fracs_flat[i].shape[0])
        peak_vals[i, :num_fracs] = fracs_flat[i][:num_fracs]

    peaks_output = peaks_output.reshape((*original_shape, 5, 3))
    peak_indices = peak_indices.reshape((*original_shape, 5))
    peak_vals = peak_vals.reshape((*original_shape, 5))

    return peaks_output, peak_indices, peak_vals


class FORCEModel(ReconstModel):
    """FORCE reconstruction model for signal matching based microstructure."""

    def __init__(
        self,
        gtab,
        *,
        simulations=None,
        penalty=1e-5,
        n_neighbors=50,
        use_posterior=False,
        posterior_beta=2000.0,
        compute_odf=False,
        verbose=False,
    ):
        r"""
        FORCE (FORward modeling for Complex microstructure Estimation) model
        :footcite:p:`Shah2025`.

        FORCE is a forward modeling paradigm that reframes how diffusion data
        is analyzed. Instead of inverting the measured signal, FORCE simulates
        a large set of biologically plausible intra-voxel fiber configurations
        and tissue compositions. It then identifies the best-matching simulation
        for each voxel by operating directly in the signal space.

        Parameters
        ----------
        gtab : GradientTable
            Gradient table.
        simulations : dict or None, optional
            Pre-computed FORCE simulations with signals and parameters.
            If None, call generate() to create simulations.
        penalty : float, optional
            Penalty weight for fiber complexity.
        n_neighbors : int, optional
            Number of neighbors for matching.
        use_posterior : bool, optional
            Use posterior averaging instead of best match.
        posterior_beta : float, optional
            Softmax temperature for posterior.
        compute_odf : bool, optional
            Compute posterior ODF maps.
        verbose : bool, optional
            Show progress bar and status messages.

        Notes
        -----
        The fit method uses the @multi_voxel_fit decorator which supports
        parallel execution. Pass `engine` and `n_jobs` kwargs to the fit method:

        Available engines: "serial", "ray", "joblib", "dask".

        References
        ----------
        .. footbibliography::
        """
        self.gtab = gtab
        self.simulations = simulations
        self.penalty = penalty
        self.n_neighbors = n_neighbors
        self.use_posterior = use_posterior
        self.posterior_beta = posterior_beta
        self.compute_odf = compute_odf
        self.verbose = verbose

        self._index = None
        self._penalty_array = None

        if simulations is not None:
            self._prepare_library()

    def generate(
        self,
        *,
        num_simulations=500000,
        output_path=None,
        num_cpus=1,
        wm_threshold=0.5,
        tortuosity=False,
        odi_range=(0.01, 0.3),
        diffusivity_config=None,
        compute_dti=True,
        compute_dki=False,
        verbose=False,
        use_cache=True,
    ):
        """Generate simulations for matching.

        When ``output_path`` is ``None`` and ``use_cache`` is ``True``,
        simulations are cached in ``~/.dipy/force_simulations/`` (or
        ``$DIPY_HOME``). A registry file (``cache_registry.json``) keeps
        track of the bvals, bvecs, diffusivity configuration and number of
        simulations for each cached file. If a cached simulation that matches
        the current gradient table (within tolerance) and diffusivity
        configuration already exists, it is loaded from disk and generation
        is skipped.

        Set ``use_cache=False`` to force regeneration even when a matching
        cached simulation exists.

        Parameters
        ----------
        num_simulations : int, optional
            Number of simulated voxels.
        output_path : str, optional
            Path to save simulations (.npz). When None, saves to
            ``~/.dipy/force_simulations/`` and uses caching.
        num_cpus : int, optional
            Number of CPU cores for parallel processing.
        wm_threshold : float, optional
            Minimum WM fraction to include fiber labels.
        tortuosity : bool, optional
            Use tortuosity constraint for perpendicular diffusivity.
        odi_range : tuple, optional
            (min, max) orientation dispersion index range.
        diffusivity_config : dict, optional
            Custom diffusivity ranges.
        compute_dti : bool, optional
            Compute DTI metrics (FA, MD, RD).
        compute_dki : bool, optional
            Compute DKI metrics (AK, RK, MK, KFA).
        verbose : bool, optional
            Enable progress output.
        use_cache : bool, optional
            Whether to use cached simulations when ``output_path`` is
            None. Set to ``False`` to always regenerate.

        Returns
        -------
        self : FORCEModel
            Model with simulations loaded.
        """

        b0_thr = getattr(self.gtab, "b0_threshold", 50)
        unique_bvals = unique_bvals_tolerance(self.gtab.bvals, tol=50)
        n_nonzero_shells = int(np.sum(unique_bvals > b0_thr))

        if compute_dki and n_nonzero_shells < 2:
            warnings.warn(
                f"compute_dki=True but only {n_nonzero_shells} non-zero "
                "b-value shell found (need at least 2). "
                "Disabling DKI computation.",
                UserWarning,
                stacklevel=2,
            )
            compute_dki = False
        elif not compute_dki and n_nonzero_shells >= 2:
            logger.info(
                f"Found {n_nonzero_shells} non-zero b-value shells. "
                "You can compute DKI metrics (AK, RK, MK, KFA) by "
                "setting compute_dki=True."
            )

        # Resolve the diffusivity config that will actually be used
        resolved_config = (
            diffusivity_config
            if diffusivity_config is not None
            else get_default_diffusivity_config()
        )

        # --- Cache logic when no explicit output_path is given ----------
        if output_path is None and use_cache:
            cache_dir = _get_force_cache_dir()
            cached = _find_cached_simulation(
                cache_dir,
                self.gtab,
                resolved_config,
                num_simulations,
            )
            if cached is not None:
                if verbose:
                    print(f"[FORCE] Loading cached simulations from {cached}")
                self.simulations = load_force_simulations(cached)
                self._prepare_library()
                return self

        # --- Generate new simulations -----------------------------------
        self.simulations = generate_force_simulations(
            self.gtab,
            num_simulations=num_simulations,
            num_cpus=num_cpus,
            wm_threshold=wm_threshold,
            tortuosity=tortuosity,
            odi_range=odi_range,
            diffusivity_config=diffusivity_config,
            compute_dti=compute_dti,
            compute_dki=compute_dki,
            verbose=verbose,
        )

        if output_path is not None:
            save_force_simulations(self.simulations, output_path)
        else:
            # Save into the .dipy cache and register.
            # filename is generated inside the lock to avoid races between
            # concurrent processes reading the same registry length.
            cache_dir = _get_force_cache_dir()
            filename_holder = {}

            def _append_and_name(registry):
                idx = len(registry)
                fname = f"force_sim_{idx}.npz"
                filename_holder["filename"] = fname
                return registry  # entry added by _register_cached_simulation

            _locked_registry_update(cache_dir, _append_and_name)
            filename = filename_holder["filename"]
            save_force_simulations(self.simulations, str(cache_dir / filename))
            _register_cached_simulation(
                cache_dir,
                self.gtab,
                resolved_config,
                num_simulations,
                filename,
            )
            if verbose:
                print(f"[FORCE] Cached simulations to {cache_dir / filename}")

        self._prepare_library()
        return self

    def load(self, input_path):
        """Load pre-computed simulations from file.

        Parameters
        ----------
        input_path : str
            Path to simulations file (.npz).

        Returns
        -------
        self : FORCEModel
            Model with simulations loaded.

        """
        self.simulations = load_force_simulations(input_path)
        self._prepare_library()
        return self

    def _prepare_library(self):
        """Prepare library for matching."""
        signals = self.simulations["signals"]

        # Normalize library signals
        lib_norm = np.linalg.norm(signals, axis=1, keepdims=True)
        lib_norm[lib_norm == 0] = 1.0
        signals_norm = np.ascontiguousarray((signals / lib_norm).astype(np.float32))

        # Build index
        self._index = create_signal_index(signals_norm)

        # Penalty array
        num_fibers = self.simulations.get(
            "num_fibers", np.zeros(len(signals), dtype=np.float32)
        )
        self._penalty_array = (self.penalty * num_fibers).astype(np.float32)

        # Prior ranges for microstructure uncertainty normalization
        d = self.simulations
        self._prior_ranges = {}
        for param in MICRO_PARAMS:
            if param in d:
                arr = d[param]
                self._prior_ranges[param] = float(arr.max() - arr.min())

    @staticmethod
    def _fetch_params_batched(lib_idx, d):
        """Vectorised parameter look-up for best-match indices.

        Parameters
        ----------
        lib_idx : ndarray (N,)
            Library indices of the best match per voxel.
        d : dict
            Simulation dictionary.

        Returns
        -------
        params : dict of ndarray
        """
        params = {
            "fa": d["fa"][lib_idx].astype(np.float32),
            "md": d["md"][lib_idx].astype(np.float32),
            "rd": d["rd"][lib_idx].astype(np.float32),
            "wm_fraction": d["wm_fraction"][lib_idx].astype(np.float32),
            "gm_fraction": d["gm_fraction"][lib_idx].astype(np.float32),
            "csf_fraction": d["csf_fraction"][lib_idx].astype(np.float32),
            "num_fibers": d["num_fibers"][lib_idx].astype(np.float32),
            "dispersion": d["dispersion"][lib_idx].astype(np.float32),
            "nd": d["nd"][lib_idx].astype(np.float32),
            "labels": d["labels"][lib_idx].astype(np.int8),
            "fracs": d["fraction_array"][lib_idx].astype(np.float32),
        }
        if "ufa_wm" in d:
            params["ufa_wm"] = d["ufa_wm"][lib_idx].astype(np.float32)
            params["ufa_voxel"] = d["ufa_voxel"][lib_idx].astype(np.float32)
        if "ak" in d:
            params["ak"] = d["ak"][lib_idx].astype(np.float32)
            params["rk"] = d["rk"][lib_idx].astype(np.float32)
            params["mk"] = d["mk"][lib_idx].astype(np.float32)
            params["kfa"] = d["kfa"][lib_idx].astype(np.float32)
        params["odf"] = d["odfs"][lib_idx].astype(np.float32) if "odfs" in d else None
        params["predicted_signal"] = d["signals"][lib_idx].astype(np.float32)
        return params

    @staticmethod
    def _posterior_params_batched(neighbors, W, d, lib_idx):
        """Vectorised posterior-averaging over neighbours.

        Parameters
        ----------
        neighbors : ndarray (N, K)
            Neighbour indices.
        W : ndarray (N, K)
            Posterior weights.
        d : dict
            Simulation dictionary.
        lib_idx : int
            Index of the exact match in the simulation

        Returns
        -------
        params : dict of ndarray
        """

        def _wavg(field):
            return np.sum(W * d[field][neighbors], axis=1).astype(np.float32)

        params = {
            "fa": _wavg("fa"),
            "md": _wavg("md"),
            "rd": _wavg("rd"),
            "wm_fraction": _wavg("wm_fraction"),
            "gm_fraction": _wavg("gm_fraction"),
            "csf_fraction": _wavg("csf_fraction"),
            "num_fibers": _wavg("num_fibers"),
            "dispersion": _wavg("dispersion"),
            "nd": _wavg("nd"),
            "labels": d["labels"][lib_idx].astype(np.int8),
            "fracs": d["fraction_array"][lib_idx].astype(np.float32),
        }
        if "ufa_wm" in d:
            params["ufa_wm"] = _wavg("ufa_wm")
            params["ufa_voxel"] = _wavg("ufa_voxel")
        if "ak" in d:
            params["ak"] = _wavg("ak")
            params["rk"] = _wavg("rk")
            params["mk"] = _wavg("mk")
            params["kfa"] = _wavg("kfa")

        # Posterior ODF
        if "odfs" in d:
            K = neighbors.shape[1]
            odf = np.zeros((neighbors.shape[0], d["odfs"].shape[1]), dtype=np.float32)
            for kk in range(K):
                odf_k = d["odfs"][neighbors[:, kk]].astype(np.float32)
                odf_k /= np.max(odf_k, axis=1, keepdims=True) + EPSILON
                odf += W[:, kk : kk + 1] * odf_k
            odf /= np.max(odf, axis=1, keepdims=True) + EPSILON
            params["odf"] = odf
        else:
            params["odf"] = None

        # Posterior mean signal
        params["predicted_signal"] = posterior_mean_signal(d["signals"], W, neighbors)
        return params

    @multi_voxel_fit(
        batched=True,
        shared_obj=("_penalty_array", "_index", "simulations"),
        chunk_size={"serial": 10_000, "ray": "auto"},
    )
    def fit(self, data, *, mask=None, **kwargs):
        """Fit model to data.

        Parameters
        ----------
        data : ndarray
            Diffusion data for a single voxel (1D) or multiple voxels (ND).
        mask : ndarray, optional
            Brain mask (for multi-voxel data).
        **kwargs : dict
            Additional arguments passed to multi_voxel_fit decorator:
            - engine : str, optional
                Parallel engine: "serial", "ray", "joblib", "dask".
            - n_jobs : int, optional
                Number of parallel jobs.
            - verbose : bool, optional
                Show progress bar.

        Returns
        -------
        fit : FORCEFit or ndarray of FORCEFit
            Fitted model for a single voxel (1-D input) or an object array
            of fitted models for a batch of voxels (2-D input).

        Notes
        -----
        This method is decorated with @multi_voxel_fit(batched=True)
        which handles multi-voxel dispatch, mask application, and aggregation
        into a MultiVoxelFit.  The method itself handles both 1-D
        (single voxel) and 2-D (batch) inputs directly.

        For parallel execution, use engine="ray", n_jobs=4 arguments in model
        fit() call.


        **Memory warning (joblib / dask engines):** When engine="joblib"
        or engine="dask", the full simulation library (including the
        signal matrix and search index, ~120-400 MB for 100k simulations) is
        pickled and sent to *every* worker chunk.  With 8 workers this can
        consume several gigabytes of RAM.  For num_simulations > ~10 000
        use engine="ray" instead, which places the library in a shared
        object store and avoids redundant copies across workers.
        """
        if self.simulations is None:
            raise RuntimeError(
                "No simulations loaded. Call generate() or provide simulations."
            )
        if self._index is None:
            raise RuntimeError(
                "Signal index is not prepared. Call _prepare_library() or "
                "reload simulations via generate() or load()."
            )

        single = data.ndim == 1
        data2d = data.reshape(1, -1) if single else data
        data2d = np.ascontiguousarray(data2d, dtype=np.float32)

        norms = np.linalg.norm(data2d, axis=1, keepdims=True).astype(np.float32)
        norms[norms == 0] = 1.0
        query_norm = np.ascontiguousarray(data2d / norms)

        D, neighbors = self._index.search(query_norm, k=self.n_neighbors)
        S = D - self._penalty_array[neighbors]

        U, A = compute_uncertainty_ambiguity(S)

        d = self.simulations
        n_vox = data2d.shape[0]
        best = np.argmax(S, axis=1)
        lib_idx = neighbors[np.arange(n_vox), best]

        W = softmax_stable(self.posterior_beta * S, axis=1)

        if self.use_posterior:
            entropy = -np.sum(W * np.log(W + EPSILON), axis=1)

            params_arrays = self._posterior_params_batched(neighbors, W, d, lib_idx)
            params_arrays["uncertainty"] = U
            params_arrays["ambiguity"] = A
            params_arrays["entropy"] = entropy.astype(np.float32)
        else:
            params_arrays = self._fetch_params_batched(lib_idx, d)
            params_arrays["uncertainty"] = U
            params_arrays["ambiguity"] = A
            params_arrays["entropy"] = np.full(n_vox, np.nan, dtype=np.float32)

        # Per-microstructure uncertainty and ambiguity from posterior density
        prior_ranges = getattr(self, "_prior_ranges", {})
        for param in MICRO_PARAMS:
            if param in d and param in prior_ranges:
                vals = d[param][neighbors].astype(np.float32)
                u, a = compute_microstructure_uncertainty_ambiguity(
                    vals, W, prior_ranges[param]
                )
                params_arrays[f"uncertainty_{param}"] = u
                params_arrays[f"ambiguity_{param}"] = a

        if kwargs.pop("_raw", False):
            return params_arrays

        fits = np.empty(n_vox, dtype=object)
        keys = list(params_arrays.keys())
        for i in range(n_vox):
            p = {}
            for k in keys:
                val = params_arrays[k]
                if val is None:
                    p[k] = None
                else:
                    v = val[i]
                    if isinstance(v, np.ndarray) and v.ndim == 0:
                        p[k] = float(v)
                    elif isinstance(v, (np.floating, np.integer)):
                        p[k] = float(v)
                    else:
                        p[k] = v
            entropy_val = p.get("entropy", 0.0)
            if entropy_val is not None and np.isnan(entropy_val):
                p["entropy"] = None
            fits[i] = FORCEFit(None, p)

        return fits[0] if single else fits


class FORCEFit(ReconstFit):
    """FORCE model fit results for a single voxel."""

    def __init__(self, model, params):
        """Initialize a FORCEFit class instance."""
        if (
            "entropy" in params
            and params["entropy"] is not None
            and np.isnan(params["entropy"])
        ):
            params["entropy"] = None
        self.model = model
        self._params = params

    @property
    def fa(self):
        """Fractional anisotropy."""
        return self._params["fa"]

    @property
    def md(self):
        """Mean diffusivity."""
        return self._params["md"]

    @property
    def rd(self):
        """Radial diffusivity."""
        return self._params["rd"]

    @property
    def wm_fraction(self):
        """White matter fraction."""
        return self._params["wm_fraction"]

    @property
    def gm_fraction(self):
        """Gray matter fraction."""
        return self._params["gm_fraction"]

    @property
    def csf_fraction(self):
        """CSF fraction."""
        return self._params["csf_fraction"]

    @property
    def num_fibers(self):
        """Number of fibers."""
        return self._params["num_fibers"]

    @property
    def dispersion(self):
        """Orientation dispersion."""
        return self._params["dispersion"]

    @property
    def nd(self):
        """Neurite density."""
        return self._params["nd"]

    @property
    def ufa_wm(self):
        """microFA in white matter."""
        return self._params.get("ufa_wm", None)

    @property
    def ufa_voxel(self):
        """Voxel-averaged microFA."""
        return self._params.get("ufa_voxel", None)

    @property
    def ak(self):
        """Axial kurtosis (DKI)."""
        return self._params.get("ak", None)

    @property
    def rk(self):
        """Radial kurtosis (DKI)."""
        return self._params.get("rk", None)

    @property
    def mk(self):
        """Mean kurtosis (DKI)."""
        return self._params.get("mk", None)

    @property
    def kfa(self):
        """Kurtosis fractional anisotropy (DKI)."""
        return self._params.get("kfa", None)

    @property
    def odf(self):
        """Orientation distribution function."""
        return self._params.get("odf", None)

    @property
    def predicted_signal(self):
        """Predicted signal from matched library entry (cleaned DWI)."""
        return self._params.get("predicted_signal", None)

    @property
    def uncertainty(self):
        """Uncertainty (IQR of penalized scores)."""
        return self._params["uncertainty"]

    @property
    def ambiguity(self):
        """Ambiguity (fraction above half-max)."""
        return self._params["ambiguity"]

    @property
    def entropy(self):
        """Entropy (posterior mode only)."""
        return self._params.get("entropy", None)

    @property
    def label(self):
        """Fiber configuration label."""
        return self._params.get("labels", None)

    @property
    def fracs(self):
        """Fiber fractions."""
        return self._params.get("fracs", None)

    @property
    def uncertainty_fa(self):
        """Per-microstructure uncertainty for FA (fraction of prior range)."""
        return self._params.get("uncertainty_fa", None)

    @property
    def ambiguity_fa(self):
        """Per-microstructure ambiguity for FA (fraction of prior range)."""
        return self._params.get("ambiguity_fa", None)

    @property
    def uncertainty_md(self):
        """Per-microstructure uncertainty for MD (fraction of prior range)."""
        return self._params.get("uncertainty_md", None)

    @property
    def ambiguity_md(self):
        """Per-microstructure ambiguity for MD (fraction of prior range)."""
        return self._params.get("ambiguity_md", None)

    @property
    def uncertainty_rd(self):
        """Per-microstructure uncertainty for RD (fraction of prior range)."""
        return self._params.get("uncertainty_rd", None)

    @property
    def ambiguity_rd(self):
        """Per-microstructure ambiguity for RD (fraction of prior range)."""
        return self._params.get("ambiguity_rd", None)

    @property
    def uncertainty_wm_fraction(self):
        """Per-microstructure uncertainty for WM fraction (fraction of prior range)."""
        return self._params.get("uncertainty_wm_fraction", None)

    @property
    def ambiguity_wm_fraction(self):
        """Per-microstructure ambiguity for WM fraction (fraction of prior range)."""
        return self._params.get("ambiguity_wm_fraction", None)

    @property
    def uncertainty_gm_fraction(self):
        """Per-microstructure uncertainty for GM fraction (fraction of prior range)."""
        return self._params.get("uncertainty_gm_fraction", None)

    @property
    def ambiguity_gm_fraction(self):
        """Per-microstructure ambiguity for GM fraction (fraction of prior range)."""
        return self._params.get("ambiguity_gm_fraction", None)

    @property
    def uncertainty_csf_fraction(self):
        """Per-microstructure uncertainty for CSF fraction (fraction of prior range)."""
        return self._params.get("uncertainty_csf_fraction", None)

    @property
    def ambiguity_csf_fraction(self):
        """Per-microstructure ambiguity for CSF fraction (fraction of prior range)."""
        return self._params.get("ambiguity_csf_fraction", None)

    @property
    def uncertainty_num_fibers(self):
        """Per-microstructure uncertainty for num_fibers (fraction of prior range)."""
        return self._params.get("uncertainty_num_fibers", None)

    @property
    def ambiguity_num_fibers(self):
        """Per-microstructure ambiguity for num_fibers (fraction of prior range)."""
        return self._params.get("ambiguity_num_fibers", None)

    @property
    def uncertainty_dispersion(self):
        """Per-microstructure uncertainty for dispersion (fraction of prior range)."""
        return self._params.get("uncertainty_dispersion", None)

    @property
    def ambiguity_dispersion(self):
        """Per-microstructure ambiguity for dispersion (fraction of prior range)."""
        return self._params.get("ambiguity_dispersion", None)

    @property
    def uncertainty_nd(self):
        """Per-microstructure uncertainty for ND (fraction of prior range)."""
        return self._params.get("uncertainty_nd", None)

    @property
    def ambiguity_nd(self):
        """Per-microstructure ambiguity for ND (fraction of prior range)."""
        return self._params.get("ambiguity_nd", None)

    @property
    def uncertainty_ufa_voxel(self):
        """Uncertainty for microFA voxel (fraction of prior range)."""
        return self._params.get("uncertainty_ufa_voxel", None)

    @property
    def ambiguity_ufa_voxel(self):
        """Per-microstructure ambiguity for microFA voxel (fraction of prior range)."""
        return self._params.get("ambiguity_ufa_voxel", None)

    @property
    def uncertainty_ak(self):
        """Per-microstructure uncertainty for AK (fraction of prior range)."""
        return self._params.get("uncertainty_ak", None)

    @property
    def ambiguity_ak(self):
        """Per-microstructure ambiguity for AK (fraction of prior range)."""
        return self._params.get("ambiguity_ak", None)

    @property
    def uncertainty_rk(self):
        """Per-microstructure uncertainty for RK (fraction of prior range)."""
        return self._params.get("uncertainty_rk", None)

    @property
    def ambiguity_rk(self):
        """Per-microstructure ambiguity for RK (fraction of prior range)."""
        return self._params.get("ambiguity_rk", None)

    @property
    def uncertainty_mk(self):
        """Per-microstructure uncertainty for MK (fraction of prior range)."""
        return self._params.get("uncertainty_mk", None)

    @property
    def ambiguity_mk(self):
        """Per-microstructure ambiguity for MK (fraction of prior range)."""
        return self._params.get("ambiguity_mk", None)

    @property
    def uncertainty_kfa(self):
        """Per-microstructure uncertainty for KFA (fraction of prior range)."""
        return self._params.get("uncertainty_kfa", None)

    @property
    def ambiguity_kfa(self):
        """Per-microstructure ambiguity for KFA (fraction of prior range)."""
        return self._params.get("ambiguity_kfa", None)


# Resolve forward reference: FORCEModel is defined before FORCEFit.
FORCEModel._fit_class = FORCEFit


def compute_entropy(weights):
    """Compute entropy of posterior weights.

    Parameters
    ----------
    weights : ndarray (N, K)
        Posterior weights for K neighbors.

    Returns
    -------
    entropy : ndarray (N,)
        Shannon entropy for each sample.
    """
    return (-np.sum(weights * np.log(weights + EPSILON), axis=1)).astype(np.float32)


def posterior_mean_signal(signals, weights, indices):
    """Compute posterior mean signal from neighbors.

    Parameters
    ----------
    signals : ndarray (N_lib, M)
        Library signals.
    weights : ndarray (N_query, K)
        Posterior weights.
    indices : ndarray (N_query, K)
        Neighbor indices.

    Returns
    -------
    mean_signal : ndarray (N_query, M)
        Posterior mean signals.
    """
    n_query = indices.shape[0]
    n_grad = signals.shape[1]
    k = indices.shape[1]

    result = np.zeros((n_query, n_grad), dtype=np.float32)
    for kk in range(k):
        result += weights[:, kk : kk + 1] * signals[indices[:, kk]]

    return result


def posterior_odf(odfs, weights, indices, n_dirs):
    """Compute posterior ODF from neighbors.

    Parameters
    ----------
    odfs : ndarray (N_lib, D)
        Library ODFs.
    weights : ndarray (N_query, K)
        Posterior weights.
    indices : ndarray (N_query, K)
        Neighbor indices.
    n_dirs : int
        Number of sphere directions.

    Returns
    -------
    odf : ndarray (N_query, D)
        Posterior mean ODFs.
    """
    n_query = indices.shape[0]
    k = indices.shape[1]

    result = np.zeros((n_query, n_dirs), dtype=np.float32)
    for kk in range(k):
        odf_k = odfs[indices[:, kk]].astype(np.float32)
        odf_k /= np.max(odf_k, axis=1, keepdims=True) + EPSILON
        result += weights[:, kk : kk + 1] * odf_k

    result /= np.max(result, axis=1, keepdims=True) + EPSILON
    return result


def force_peaks(fitted_object, *, mask=None, sh_order=8):
    """Create a PeaksAndMetrics object from a FORCEFit or MultiVoxelFit.

    Parameters
    ----------
    fitted_object : FORCEFit or MultiVoxelFit
        The result of model.fit().
    mask : ndarray, optional
        Optional brain mask.
    sh_order : int, optional
        Spherical harmonics order for the coefficients.
    """
    from dipy.sims.force import default_sphere

    labels = fitted_object.label
    fracs = fitted_object.fracs
    odf = fitted_object.odf

    is_multi_voxel = labels.ndim > 1

    if not is_multi_voxel:
        # Single Voxel Case
        p_out, p_ind, p_val = postprocess_peaks(
            labels[None, :], default_sphere, fracs[None, :]
        )
        res_dirs, res_inds, res_vals = p_out[0], p_ind[0], p_val[0]
        res_sh = (
            sf_to_sh(odf, default_sphere, sh_order=sh_order)
            if odf is not None
            else None
        )
    else:
        # Multi-Voxel / CLI Case
        original_shape = labels.shape[:-1]  # (X, Y, Z)

        if mask is not None:
            labels_to_proc = labels[mask]
            fracs_to_proc = fracs[mask]
        else:
            labels_to_proc = labels.reshape(-1, labels.shape[-1])
            fracs_to_proc = fracs.reshape(-1, fracs.shape[-1])

        p_out, p_ind, p_val = postprocess_peaks(
            labels_to_proc, default_sphere, fracs_to_proc
        )

        res_dirs = np.zeros((*original_shape, 5, 3), dtype=np.float32)
        res_inds = np.full((*original_shape, 5), -1, dtype=np.int32)
        res_vals = np.zeros((*original_shape, 5), dtype=np.float32)

        if mask is not None:
            res_dirs[mask] = p_out
            res_inds[mask] = p_ind
            res_vals[mask] = p_val
        else:
            res_dirs = p_out.reshape((*original_shape, 5, 3))
            res_inds = p_ind.reshape((*original_shape, 5))
            res_vals = p_val.reshape((*original_shape, 5))

        res_sh = None
        if odf is not None and np.issubdtype(
            getattr(odf, "dtype", type(None)), np.floating
        ):
            v_max = np.max(odf, axis=-1, keepdims=True)
            v_min = np.min(odf, axis=-1, keepdims=True)
            mask = v_max > 1.0
            denom = (v_max - v_min) + 1e-12
            normalized_odf = (odf - v_min) / denom
            odf = np.where(mask, normalized_odf, odf)
            res_sh = sf_to_sh(odf, default_sphere, sh_order=sh_order)
    peaks = PeaksAndMetrics()
    peaks.peak_dirs = res_dirs
    peaks.peak_values = res_vals
    peaks.peak_indices = res_inds
    peaks.shm_coeff = res_sh
    peaks.sphere = default_sphere

    return peaks



from unittest.mock import patch

import numpy as np
import pytest

from dipy.direction.peaks import PeaksAndMetrics
from dipy.testing.decorators import set_random_number_generator, use_xvfb
from dipy.utils.optpkg import optional_package
from dipy.viz.horizon.tab.base import HorizonTab, HorizonUIElement
from dipy.viz.horizon.tab.peak import PeaksTab
from dipy.viz.horizon.visualizer.peak import PeaksVisualizer

fury, has_fury, setup_module = optional_package(
    "fury", min_version="0.10.0", max_version="1.0.0"
)
skip_it = use_xvfb == "skip" or not has_fury


@pytest.fixture
@set_random_number_generator()
def peak_actor(rng):
    """Fixture to create a Peaks Actor."""
    peak_dirs = 255 * rng.random((5, 5, 5, 5, 3))
    pam = PeaksAndMetrics()
    pam.peak_dirs = peak_dirs
    pam.affine = np.eye(4)

    peak_viz = PeaksVisualizer((pam.peak_dirs, pam.affine), False, "mock_peaks")
    return peak_viz.actors[0]


@pytest.fixture
def slider():
    """Fixture to create a mock slider."""

    class MockSlider:
        def __init__(self):
            self.value = 0
            self.left_disk_value = 0
            self.right_disk_value = 100
            self.position = (0, 0)
            self.visibility = True

        def set_position(self, x, y):
            self.position = (x, y)

        def set_visibility(self, visibility):
            self.visibility = visibility

    return MockSlider()


@pytest.mark.skipif(skip_it, reason="Needs xvfb")
def test_init_params(peak_actor):
    """Test initialization of PeaksTab."""
    tab = PeaksTab(peak_actor, "Peaks 1", "mock_peaks")
    assert isinstance(tab, PeaksTab)
    assert tab._actor == peak_actor
    assert tab.name == "Peaks 1"
    assert tab._file_name == "mock_peaks"


@pytest.mark.skipif(skip_it, reason="Needs xvfb")
def test_init_no_file_name(peak_actor):
    """Test initialization of PeaksTab without file name."""
    tab = PeaksTab(peak_actor, "Peaks 1", None)
    assert isinstance(tab, PeaksTab)
    assert tab._actor == peak_actor
    assert tab.name == "Peaks 1"
    assert tab._file_name == "Peaks 1"


@pytest.mark.skipif(skip_it, reason="Needs xvfb")
def test_init_ui_element_creation(peak_actor):
    """Test UI element creation in PeaksTab."""
    tab = PeaksTab(peak_actor, "Peaks 1", "mock_peaks")
    assert tab._opacity is not None
    assert tab._opacity_label is not None
    assert tab._slice_x is not None
    assert tab._slice_y is not None
    assert tab._slice_z is not None
    assert tab._slice_x_label is not None
    assert tab._slice_y_label is not None
    assert tab._slice_z_label is not None
    assert tab._slice_x_toggle is not None
    assert tab._slice_y_toggle is not None
    assert tab._slice_z_toggle is not None
    assert tab._range_x is not None
    assert tab._range_y is not None
    assert tab._range_z is not None
    assert tab._range_x_label is not None
    assert tab._range_y_label is not None
    assert tab._range_z_label is not None
    assert tab._view_mode_label is not None
    assert tab._view_mode_toggler is not None
    assert tab._file_label is not None
    assert tab._file_name_label is not None


@pytest.mark.skipif(skip_it, reason="Needs xvfb")
def test_update_ui_elements_position(peak_actor):
    """Test updating UI elements position in PeaksTab."""
    tab = PeaksTab(peak_actor, "Peaks 1", "mock_peaks")

    tab.build(0)

    assert tab._tab_id == 0

    x_pos = 0.02
    assert tab._actor_toggle.position == (x_pos, 0.85)
    assert tab._slice_x_toggle.position == (x_pos, 0.62)
    assert tab._slice_y_toggle.position == (x_pos, 0.38)
    assert tab._slice_z_toggle.position == (x_pos, 0.15)

    x_pos = 0.04
    assert tab._opacity_label.position == (x_pos, 0.85)
    assert tab._slice_x_label.position == (x_pos, 0.62)
    assert tab._slice_y_label.position == (x_pos, 0.38)
    assert tab._slice_z_label.position == (x_pos, 0.15)
    assert tab._range_x_label.position == (x_pos, 0.62)
    assert tab._range_y_label.position == (x_pos, 0.38)
    assert tab._range_z_label.position == (x_pos, 0.15)

    x_pos = 0.10
    assert tab._opacity.position == (x_pos, 0.85)
    assert tab._slice_x.position == (x_pos, 0.62)
    assert tab._slice_y.position == (x_pos, 0.38)
    assert tab._slice_z.position == (x_pos, 0.15)

    x_pos = 0.105
    assert tab._range_x.position == (x_pos, 0.66)
    assert tab._range_y.position == (x_pos, 0.42)
    assert tab._range_z.position == (x_pos, 0.19)

    x_pos = 0.52
    assert tab._view_mode_label.position == (x_pos, 0.85)
    assert tab._file_label.position == (x_pos, 0.28)

    x_pos = 0.62
    assert tab._view_mode_toggler.position == (x_pos, 0.80)
    assert tab._file_name_label.position == (x_pos, 0.28)


@pytest.mark.skipif(skip_it, reason="Needs xvfb")
def test_update_slices(peak_actor):
    """Test updating slices in PeaksTab when synced."""
    tab = PeaksTab(peak_actor, "Peaks 1", "mock_peaks")
    tab.build(0)

    tab.update_slices(3, 2, 2)

    assert tab._slice_x.obj.value == 3
    assert tab._slice_y.obj.value == 2
    assert tab._slice_z.obj.value == 2


@patch.object(HorizonTab, "show")
@patch.object(HorizonTab, "hide")
@pytest.mark.skipif(skip_it, reason="Needs xvfb")
def test_on_tab_selected(_m_show, _m_hide, peak_actor):
    """Test on_tab_selected method in PeaksTab."""
    tab = PeaksTab(peak_actor, "Peaks 1", "mock_peaks")
    tab.build(0)

    tab._view_mode_toggler.obj.checked_labels = ["Cross section"]

    with patch.object(PeaksTab, "_show_cross_section") as mock_show_cross_section:
        tab.on_tab_selected()
        mock_show_cross_section.assert_called_once()

    tab._view_mode_toggler.obj.checked_labels = ["Range"]

    with patch.object(PeaksTab, "_show_range") as mock_show_range:
        tab.on_tab_selected()
        mock_show_range.assert_called_once()


@pytest.mark.skipif(skip_it, reason="Needs xvfb")
def test_change_opcity(peak_actor, slider):
    """Test change_opacity method in PeaksTab."""
    tab = PeaksTab(peak_actor, "Peaks 1", "mock_peaks")
    tab.build(0)

    slider.value = 0.5
    tab._change_opacity(slider)
    assert tab._actor.global_opacity == 0.5


@pytest.mark.skipif(skip_it, reason="Needs xvfb")
def test_change_range(peak_actor, slider):
    """Test change_range method in PeaksTab."""
    tab = PeaksTab(peak_actor, "Peaks 1", "mock_peaks")
    tab.build(0)

    slider.left_disk_value = 10
    slider.right_disk_value = 90

    selected_range = HorizonUIElement(True, (0, 0), None)
    tab._change_range(slider, selected_range)

    assert selected_range.selected_value == (10, 90)


@pytest.mark.skipif(skip_it, reason="Needs xvfb")
def test_change_slice(peak_actor, slider):
    """Test change_slice method in PeaksTab."""
    tab = PeaksTab(peak_actor, "Peaks 1", "mock_peaks")
    tab.build(0)

    slider.value = 5
    selected_slice = HorizonUIElement(True, 0, None)
    tab._view_mode_toggler.obj.checked_labels = ["Cross section"]

    with patch.object(HorizonTab, "on_slice_change") as mock_sync:
        tab._change_slice(slider, selected_slice, sync_slice=True)
        mock_sync.assert_not_called()
        assert selected_slice.selected_value == 5

        slider.value = 10
        tab._change_slice(slider, selected_slice)
        mock_sync.assert_called_once()
        assert selected_slice.selected_value == 10


@pytest.mark.skipif(skip_it, reason="Needs xvfb")
def test_update_slice_visibility(peak_actor, slider):
    """Test update_slice_visibility method in PeaksTab."""
    tab = PeaksTab(peak_actor, "Peaks 1", "mock_peaks")
    tab.build(0)

    selected_slice = HorizonUIElement(False, 0, slider)
    tab._update_slice_visibility(tab._slice_x_toggle.obj, selected_slice)
    assert selected_slice.visibility is True
    assert selected_slice.obj.visibility is True


@pytest.mark.skipif(skip_it, reason="Needs xvfb")
def test_toggle_actors(peak_actor):
    """Test toggle_actors method in PeaksTab."""
    tab = PeaksTab(peak_actor, "Peaks 1", "mock_peaks")
    tab.build(0)

    tab._actor_toggle.obj.checked_labels = [""]

    with patch.object(HorizonTab, "show") as mock_show:
        tab._toggle_actors(tab._actor_toggle.obj)
        mock_show.assert_called_with(*tab.actors)
        assert mock_show.call_count == 2

    tab._actor_toggle.obj.checked_labels = []

    with patch.object(HorizonTab, "hide") as mock_hide:
        tab._toggle_actors(tab._actor_toggle.obj)
        mock_hide.assert_called_with(*tab.actors)
        assert mock_hide.call_count == 2



# Header file for dkispeed.pyx
# Provides declarations for Cython-optimized DKI functions

cimport numpy as cnp

# Single-value helper functions (nogil compatible)
cdef bint positive_evals_single(double L1, double L2, double L3, 
                                 double er=*) noexcept nogil

cdef double carlson_rf_single(double x, double y, double z, 
                               double errtol=*) noexcept nogil

cdef double carlson_rd_single(double x, double y, double z, 
                               double errtol=*) noexcept nogil

cdef double F1m_single(double a, double b, double c, 
                        double er=*) noexcept nogil

cdef double F2m_single(double a, double b, double c, 
                        double er=*) noexcept nogil

cdef double G1m_single(double a, double b, double c,
                        double er=*) noexcept nogil

cdef double G2m_single(double a, double b, double c,
                        double er=*) noexcept nogil

cdef double Wrotate_element_single(double[:] kt, int indi, int indj, 
                                    int indk, int indl, 
                                    double[:, :] B) noexcept nogil



# cython: profile=True
# cython: embedsignature=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
"""
Optimized Cython routines for Diffusion Kurtosis Imaging (DKI).

This module contains Cython-optimized versions of computationally intensive
DKI functions, particularly the mean kurtosis calculation.
"""

cimport cython
import numpy as np
cimport numpy as cnp
from libc.math cimport sqrt, fabs, atan, log

cnp.import_array()


cdef inline double arctanh_c(double x) noexcept nogil:
    """Compute inverse hyperbolic tangent."""
    return 0.5 * log((1.0 + x) / (1.0 - x))

# Defining keys to select a kurtosis tensor element with indexes (i, j, k, l)
# on a kt vector that contains only the 15 independent elements of the kurtosis
# tensor.
cdef int[16] ind_ele_arr = [0, 0, 3, 4, 9, 5, 12, 7, 8, 10, 9, 11, 13, 13, 14, 11]

# Map from key = (i+1)*(j+1)*(k+1)*(l+1) to index in 15-element kt array
cdef inline int get_kt_index(int key) noexcept nogil:
    """Get index into 15-element kt array from key."""
    if key == 1:
        return 0
    elif key == 16:
        return 1
    elif key == 81:
        return 2
    elif key == 2:
        return 3
    elif key == 3:
        return 4
    elif key == 8:
        return 5
    elif key == 24:
        return 6
    elif key == 27:
        return 7
    elif key == 54:
        return 8
    elif key == 4:
        return 9
    elif key == 9:
        return 10
    elif key == 36:
        return 11
    elif key == 6:
        return 12
    elif key == 12:
        return 13
    elif key == 18:
        return 14
    else:
        return 0



cdef inline bint positive_evals_single(double L1, double L2, double L3, 
                                        double er=2e-7) noexcept nogil:
    """Check if all eigenvalues are significantly larger than zero.

    Parameters
    ----------
    L1 : double
        First eigenvalue of the diffusion tensor.
    L2 : double
        Second eigenvalue of the diffusion tensor.
    L3 : double
        Third eigenvalue of the diffusion tensor.
    er : double, optional
        Threshold below which an eigenvalue is considered non-positive.

    Returns
    -------
    result : bint
        True if all three eigenvalues exceed ``er``, False otherwise.
    """
    return L1 > er and L2 > er and L3 > er



cdef double carlson_rf_single(double x, double y, double z, 
                               double errtol=3e-4) noexcept nogil:
    """Compute Carlson's incomplete elliptic integral of the first kind
    for a single set of values.
    
    Parameters
    ----------
    x, y, z : double
        Independent variables of the integral.
    errtol : double
        Error tolerance.
    
    Returns
    -------
    RF : double
        Value of the incomplete first order elliptic integral.
    """
    cdef:
        double xn = x
        double yn = y
        double zn = z
        double An = (xn + yn + zn) / 3.0
        double Q, xnroot, ynroot, znroot, lamda
        double X, Y, Z, E2, E3, RF
        double max_diff
        int n = 0
    
    # Compute Q
    max_diff = fabs(An - xn)
    if fabs(An - yn) > max_diff:
        max_diff = fabs(An - yn)
    if fabs(An - zn) > max_diff:
        max_diff = fabs(An - zn)
    Q = (3.0 * errtol) ** (-1.0 / 6.0) * max_diff
    
    # Convergence loop
    while (4.0 ** (-n)) * Q > fabs(An):
        xnroot = sqrt(xn)
        ynroot = sqrt(yn)
        znroot = sqrt(zn)
        lamda = xnroot * (ynroot + znroot) + ynroot * znroot
        n = n + 1
        xn = (xn + lamda) * 0.25
        yn = (yn + lamda) * 0.25
        zn = (zn + lamda) * 0.25
        An = (An + lamda) * 0.25
    
    # Post convergence calculation
    X = 1.0 - xn / An
    Y = 1.0 - yn / An
    Z = -X - Y
    E2 = X * Y - Z * Z
    E3 = X * Y * Z
    RF = (An ** (-0.5)) * (1.0 - E2 / 10.0 + E3 / 14.0 + 
                           (E2 * E2) / 24.0 - 3.0 / 44.0 * E2 * E3)
    
    return RF



cdef double carlson_rd_single(double x, double y, double z, 
                               double errtol=1e-4) noexcept nogil:
    """Compute Carlson's incomplete elliptic integral of the second kind
    for a single set of values.
    
    Parameters
    ----------
    x, y, z : double
        Independent variables of the integral.
    errtol : double
        Error tolerance.
    
    Returns
    -------
    RD : double
        Value of the incomplete second order elliptic integral.
    """
    cdef:
        double xn = x
        double yn = y
        double zn = z
        double A0 = (xn + yn + 3.0 * zn) / 5.0
        double An = A0
        double Q, xnroot, ynroot, znroot, lamda
        double sum_term = 0.0
        double X, Y, Z, E2, E3, E4, E5, RD
        double max_diff
        double power_4n
        int n = 0
    
    # Compute Q
    max_diff = fabs(An - xn)
    if fabs(An - yn) > max_diff:
        max_diff = fabs(An - yn)
    if fabs(An - zn) > max_diff:
        max_diff = fabs(An - zn)
    Q = (errtol / 4.0) ** (-1.0 / 6.0) * max_diff
    
    # Convergence loop
    while (4.0 ** (-n)) * Q > fabs(An):
        xnroot = sqrt(xn)
        ynroot = sqrt(yn)
        znroot = sqrt(zn)
        lamda = xnroot * (ynroot + znroot) + ynroot * znroot
        sum_term = sum_term + (4.0 ** (-n)) / (znroot * (zn + lamda))
        n = n + 1
        xn = (xn + lamda) * 0.25
        yn = (yn + lamda) * 0.25
        zn = (zn + lamda) * 0.25
        An = (An + lamda) * 0.25
    
    # Post convergence calculation
    power_4n = 4.0 ** n
    X = (A0 - x) / (power_4n * An)
    Y = (A0 - y) / (power_4n * An)
    Z = -(X + Y) / 3.0
    E2 = X * Y - 6.0 * Z * Z
    E3 = (3.0 * X * Y - 8.0 * Z * Z) * Z
    E4 = 3.0 * (X * Y - Z * Z) * Z * Z
    E5 = X * Y * Z * Z * Z
    
    RD = ((4.0 ** (-n)) * (An ** (-1.5)) * 
          (1.0 - 3.0 / 14.0 * E2 + 1.0 / 6.0 * E3 + 
           9.0 / 88.0 * (E2 * E2) - 3.0 / 22.0 * E4 - 
           9.0 / 52.0 * E2 * E3 + 3.0 / 26.0 * E5) + 
          3.0 * sum_term)
    
    return RD



cdef double F2m_single(double a, double b, double c, double er=2.5e-2) noexcept nogil:
    """Compute function F2 for a single voxel.

    Helper function required to compute the analytical solution of Mean
    Kurtosis.

    Parameters
    ----------
    a : double
        First diffusion tensor eigenvalue.
    b : double
        Second diffusion tensor eigenvalue.
    c : double
        Third diffusion tensor eigenvalue.
    er : double, optional
        Relative tolerance for detecting degenerate eigenvalue pairs.

    Returns
    -------
    F2 : double
        Value of the F2 function. Returns 0.0 when eigenvalues are
        non-positive.
    """
    cdef:
        double L1, L2, L3, RF, RD, F2
        double x, alpha
    
    # Check if eigenvalues are positive
    if not positive_evals_single(a, b, c):
        return 0.0
    
    # Check for singularity b==c
    if fabs(b - c) < b * er:
        # Check for singularity a==b and a==c (isotropic case)
        if fabs(a - b) < b * er:
            return 6.0 / 15.0
        
        # Singularity b==c but a!=b
        L1 = a
        L3 = (c + b) / 2.0
        
        # Compute alpha
        x = 1.0 - (L1 / L3)
        if x > 0:
            alpha = 1.0 / sqrt(x) * arctanh_c(sqrt(x))
        else:
            alpha = 1.0 / sqrt(-x) * atan(sqrt(-x))
        
        F2 = (6.0 * ((L1 + 2.0 * L3) ** 2) / 
              (144.0 * L3 * L3 * (L1 - L3) * (L1 - L3)) * 
              (L3 * (L1 + 2.0 * L3) + L1 * (L1 - 4.0 * L3) * alpha))
        return F2
    
    # Non-singular case b!=c
    L1 = a
    L2 = b
    L3 = c
    RF = carlson_rf_single(L1 / L2, L1 / L3, 1.0)
    RD = carlson_rd_single(L1 / L2, L1 / L3, 1.0)
    
    F2 = (((L1 + L2 + L3) ** 2) / (3.0 * (L2 - L3) * (L2 - L3)) * 
          (((L2 + L3) / (sqrt(L2 * L3))) * RF + 
           ((2.0 * L1 - L2 - L3) / (3.0 * sqrt(L2 * L3))) * RD - 2.0))
    
    return F2



cdef double F1m_single(double a, double b, double c, double er=2.5e-2) noexcept nogil:
    """Compute function F1 for a single voxel.

    Helper function required to compute the analytical solution of Mean
    Kurtosis.

    Parameters
    ----------
    a : double
        First diffusion tensor eigenvalue.
    b : double
        Second diffusion tensor eigenvalue.
    c : double
        Third diffusion tensor eigenvalue.
    er : double, optional
        Relative tolerance for detecting degenerate eigenvalue pairs.

    Returns
    -------
    F1 : double
        Value of the F1 function. Returns 0.0 when eigenvalues are
        non-positive, or 1/5 in the fully isotropic case.
    """
    cdef:
        double L1, L2, L3, RF, RD, F1
    
    # Check if eigenvalues are positive
    if not positive_evals_single(a, b, c):
        return 0.0
    
    # Check for singularity a==b and a==c (isotropic case)
    if fabs(a - b) < a * er and fabs(a - c) < a * er:
        return 1.0 / 5.0
    
    # Check for singularity a==b
    if fabs(a - b) < a * er and fabs(a - c) >= a * er:
        L1 = (a + b) / 2.0
        L3 = c
        return F2m_single(L3, L1, L1) / 2.0
    
    # Check for singularity a==c
    if fabs(a - c) < a * er and fabs(a - b) >= a * er:
        L1 = (a + c) / 2.0
        L2 = b
        return F2m_single(L2, L1, L1) / 2.0
    
    # Non-singular case a!=b and a!=c
    L1 = a
    L2 = b
    L3 = c
    RF = carlson_rf_single(L1 / L2, L1 / L3, 1.0)
    RD = carlson_rd_single(L1 / L2, L1 / L3, 1.0)
    
    F1 = (((L1 + L2 + L3) ** 2) / (18.0 * (L1 - L2) * (L1 - L3)) * 
          ((sqrt(L2 * L3) / L1) * RF + 
           ((3.0 * L1 * L1 - L1 * L2 - L1 * L3 - L2 * L3) / 
            (3.0 * L1 * sqrt(L2 * L3))) * RD - 1.0))
    
    return F1



cdef double G1m_single(double a, double b, double c, double er=2.5e-2) noexcept nogil:
    """Compute function G1 for a single voxel.

    Helper function required to compute the analytical solution of Radial
    Kurtosis.

    Parameters
    ----------
    a : double
        First diffusion tensor eigenvalue (axial).
    b : double
        Second diffusion tensor eigenvalue (radial).
    c : double
        Third diffusion tensor eigenvalue (radial).
    er : double, optional
        Relative tolerance for detecting the singularity b == c.

    Returns
    -------
    G1 : double
        Value of the G1 function. Returns 0.0 when eigenvalues are
        non-positive.
    """
    cdef double L1, L2, G1

    if not positive_evals_single(a, b, c):
        return 0.0

    if fabs(b - c) >= b * er:
        G1 = ((a + b + c) * (a + b + c) /
              (18.0 * b * (b - c) * (b - c)) *
              (2.0 * b + (c * c - 3.0 * b * c) / sqrt(b * c)))
        return G1

    # Singularity b==c
    L1 = a
    L2 = b
    G1 = (L1 + 2.0 * L2) * (L1 + 2.0 * L2) / (24.0 * L2 * L2)
    return G1



cdef double G2m_single(double a, double b, double c, double er=2.5e-2) noexcept nogil:
    """Compute function G2 for a single voxel.

    Helper function required to compute the analytical solution of Radial
    Kurtosis.

    Parameters
    ----------
    a : double
        First diffusion tensor eigenvalue (axial).
    b : double
        Second diffusion tensor eigenvalue (radial).
    c : double
        Third diffusion tensor eigenvalue (radial).
    er : double, optional
        Relative tolerance for detecting the singularity b == c.

    Returns
    -------
    G2 : double
        Value of the G2 function. Returns 0.0 when eigenvalues are
        non-positive.
    """
    cdef double L1, L2, G2

    if not positive_evals_single(a, b, c):
        return 0.0

    if fabs(b - c) >= b * er:
        G2 = ((a + b + c) * (a + b + c) /
              (3.0 * (b - c) * (b - c)) *
              ((b + c) / sqrt(b * c) - 2.0))
        return G2

    # Singularity b==c
    L1 = a
    L2 = b
    G2 = (L1 + 2.0 * L2) * (L1 + 2.0 * L2) / (12.0 * L2 * L2)
    return G2



cdef double Wrotate_element_single(double[:] kt, int indi, int indj, 
                                    int indk, int indl, 
                                    double[:, :] B) noexcept nogil:
    """Compute a single rotated kurtosis tensor element.
    
    Parameters
    ----------
    kt : array (15,)
        15 independent elements of the kurtosis tensor.
    indi, indj, indk, indl : int
        Indices of the rotated tensor element (0, 1, or 2 for x, y, z).
    B : array (3, 3)
        Eigenvector matrix (columns are eigenvectors).
    
    Returns
    -------
    Wre : double
        Rotated kurtosis tensor element.
    """
    cdef:
        double Wre = 0.0
        int il, jl, kl, ll, key, idx
        double multiplyB
    
    for il in range(3):
        for jl in range(3):
            for kl in range(3):
                for ll in range(3):
                    key = (il + 1) * (jl + 1) * (kl + 1) * (ll + 1)
                    idx = get_kt_index(key)
                    multiplyB = (B[il, indi] * B[jl, indj] * 
                                 B[kl, indk] * B[ll, indl])
                    Wre = Wre + multiplyB * kt[idx]
    
    return Wre



def mean_kurtosis_analytical(double[:, :] dki_params_flat,
                              double min_kurtosis=-3.0/7.0,
                              double max_kurtosis=3.0):
    """Compute mean kurtosis using the analytical solution.

    This is the Cython-optimized version of the analytical mean kurtosis
    calculation.

    Parameters
    ----------
    dki_params_flat : ndarray (n, 27)
        Flattened DKI parameters array.
    min_kurtosis : float, optional
        Minimum kurtosis value for clipping.
    max_kurtosis : float, optional
        Maximum kurtosis value for clipping.

    Returns
    -------
    MK : ndarray (n,)
        Mean kurtosis values.
    """
    cdef:
        cnp.npy_intp n_voxels = dki_params_flat.shape[0]
        cnp.npy_intp v
        double[:] MK = np.zeros(n_voxels, dtype=np.float64)
        double[:] evals = np.zeros(3, dtype=np.float64)
        double[:, :] evecs = np.zeros((3, 3), dtype=np.float64)
        double[:] kt = np.zeros(15, dtype=np.float64)
        double L1, L2, L3
        double Wxxxx, Wyyyy, Wzzzz, Wxxyy, Wxxzz, Wyyzz
        double F1_1, F1_2, F1_3, F2_1, F2_2, F2_3
        double mk_val
        int i, j
    
    with nogil:
        for v in range(n_voxels):
            # Extract eigenvalues
            L1 = dki_params_flat[v, 0]
            L2 = dki_params_flat[v, 1]
            L3 = dki_params_flat[v, 2]
            
            # Check if eigenvalues are valid
            if not positive_evals_single(L1, L2, L3):
                MK[v] = 0.0
                continue
            
            # Extract eigenvectors (stored as 3 rows of 3 elements each)
            for i in range(3):
                for j in range(3):
                    evecs[i, j] = dki_params_flat[v, 3 + i * 3 + j]
            
            # Extract kurtosis tensor elements
            for i in range(15):
                kt[i] = dki_params_flat[v, 12 + i]
            
            # Rotate kurtosis tensor elements
            Wxxxx = Wrotate_element_single(kt, 0, 0, 0, 0, evecs)
            Wyyyy = Wrotate_element_single(kt, 1, 1, 1, 1, evecs)
            Wzzzz = Wrotate_element_single(kt, 2, 2, 2, 2, evecs)
            Wxxyy = Wrotate_element_single(kt, 0, 0, 1, 1, evecs)
            Wxxzz = Wrotate_element_single(kt, 0, 0, 2, 2, evecs)
            Wyyzz = Wrotate_element_single(kt, 1, 1, 2, 2, evecs)
            
            # Compute F1 and F2 functions
            F1_1 = F1m_single(L1, L2, L3)
            F1_2 = F1m_single(L2, L1, L3)
            F1_3 = F1m_single(L3, L2, L1)
            F2_1 = F2m_single(L1, L2, L3)
            F2_2 = F2m_single(L2, L1, L3)
            F2_3 = F2m_single(L3, L2, L1)
            
            # Compute MK
            mk_val = (F1_1 * Wxxxx + F1_2 * Wyyyy + F1_3 * Wzzzz +
                      F2_1 * Wyyzz + F2_2 * Wxxzz + F2_3 * Wxxyy)
            
            # Apply clipping
            if mk_val < min_kurtosis:
                mk_val = min_kurtosis
            if mk_val > max_kurtosis:
                mk_val = max_kurtosis
            
            MK[v] = mk_val
    
    return np.asarray(MK)



def axial_kurtosis_analytical(double[:, :] dki_params_flat,
                               double min_kurtosis=-3.0/7.0,
                               double max_kurtosis=10.0):
    """Compute axial kurtosis using the analytical solution.

    Parameters
    ----------
    dki_params_flat : ndarray (n, 27)
        Flattened DKI parameters array.
    min_kurtosis : float, optional
        Minimum kurtosis value for clipping.
    max_kurtosis : float, optional
        Maximum kurtosis value for clipping.

    Returns
    -------
    AK : ndarray (n,)
        Axial kurtosis values.
    """
    cdef:
        cnp.npy_intp n_voxels = dki_params_flat.shape[0]
        cnp.npy_intp v
        double[:] AK = np.zeros(n_voxels, dtype=np.float64)
        double[:, :] evecs = np.zeros((3, 3), dtype=np.float64)
        double[:] kt = np.zeros(15, dtype=np.float64)
        double L1, L2, L3, md, Wxxxx, ak_val
        int i, j

    with nogil:
        for v in range(n_voxels):
            L1 = dki_params_flat[v, 0]
            L2 = dki_params_flat[v, 1]
            L3 = dki_params_flat[v, 2]

            if not positive_evals_single(L1, L2, L3):
                AK[v] = 0.0
                continue

            for i in range(3):
                for j in range(3):
                    evecs[i, j] = dki_params_flat[v, 3 + i * 3 + j]

            for i in range(15):
                kt[i] = dki_params_flat[v, 12 + i]

            Wxxxx = Wrotate_element_single(kt, 0, 0, 0, 0, evecs)

            md = (L1 + L2 + L3) / 3.0
            ak_val = Wxxxx * (md * md) / (L1 * L1)

            if ak_val < min_kurtosis:
                ak_val = min_kurtosis
            if ak_val > max_kurtosis:
                ak_val = max_kurtosis

            AK[v] = ak_val

    return np.asarray(AK)



def radial_kurtosis_analytical(double[:, :] dki_params_flat,
                                double min_kurtosis=-3.0/7.0,
                                double max_kurtosis=10.0):
    """Compute radial kurtosis using the analytical solution.

    Parameters
    ----------
    dki_params_flat : ndarray (n, 27)
        Flattened DKI parameters array.
    min_kurtosis : float, optional
        Minimum kurtosis value for clipping.
    max_kurtosis : float, optional
        Maximum kurtosis value for clipping.

    Returns
    -------
    RK : ndarray (n,)
        Radial kurtosis values.
    """
    cdef:
        cnp.npy_intp n_voxels = dki_params_flat.shape[0]
        cnp.npy_intp v
        double[:] RK = np.zeros(n_voxels, dtype=np.float64)
        double[:, :] evecs = np.zeros((3, 3), dtype=np.float64)
        double[:] kt = np.zeros(15, dtype=np.float64)
        double L1, L2, L3
        double Wyyyy, Wzzzz, Wyyzz
        double rk_val
        int i, j

    with nogil:
        for v in range(n_voxels):
            L1 = dki_params_flat[v, 0]
            L2 = dki_params_flat[v, 1]
            L3 = dki_params_flat[v, 2]

            if not positive_evals_single(L1, L2, L3):
                RK[v] = 0.0
                continue

            for i in range(3):
                for j in range(3):
                    evecs[i, j] = dki_params_flat[v, 3 + i * 3 + j]

            for i in range(15):
                kt[i] = dki_params_flat[v, 12 + i]

            Wyyyy = Wrotate_element_single(kt, 1, 1, 1, 1, evecs)
            Wzzzz = Wrotate_element_single(kt, 2, 2, 2, 2, evecs)
            Wyyzz = Wrotate_element_single(kt, 1, 1, 2, 2, evecs)

            rk_val = (G1m_single(L1, L2, L3) * Wyyyy +
                      G1m_single(L1, L3, L2) * Wzzzz +
                      G2m_single(L1, L2, L3) * Wyyzz)

            if rk_val < min_kurtosis:
                rk_val = min_kurtosis
            if rk_val > max_kurtosis:
                rk_val = max_kurtosis

            RK[v] = rk_val

    return np.asarray(RK)



def kurtosis_fractional_anisotropy_c(double[:, :] dki_params_flat):
    """Compute kurtosis fractional anisotropy (KFA).

    Parameters
    ----------
    dki_params_flat : ndarray (n, 27)
        Flattened DKI parameters array.

    Returns
    -------
    KFA : ndarray (n,)
        Kurtosis fractional anisotropy values.
    """
    cdef:
        cnp.npy_intp n_voxels = dki_params_flat.shape[0]
        cnp.npy_intp v
        double[:] KFA = np.zeros(n_voxels, dtype=np.float64)
        double Wxxxx, Wyyyy, Wzzzz
        double Wxxxy, Wxxxz, Wxyyy, Wyyyz, Wxzzz, Wyzzz
        double Wxxyy, Wxxzz, Wyyzz
        double Wxxyz, Wxyyz, Wxyzz
        double W, A, B

    with nogil:
        for v in range(n_voxels):
            Wxxxx = dki_params_flat[v, 12]
            Wyyyy = dki_params_flat[v, 13]
            Wzzzz = dki_params_flat[v, 14]
            Wxxxy = dki_params_flat[v, 15]
            Wxxxz = dki_params_flat[v, 16]
            Wxyyy = dki_params_flat[v, 17]
            Wyyyz = dki_params_flat[v, 18]
            Wxzzz = dki_params_flat[v, 19]
            Wyzzz = dki_params_flat[v, 20]
            Wxxyy = dki_params_flat[v, 21]
            Wxxzz = dki_params_flat[v, 22]
            Wyyzz = dki_params_flat[v, 23]
            Wxxyz = dki_params_flat[v, 24]
            Wxyyz = dki_params_flat[v, 25]
            Wxyzz = dki_params_flat[v, 26]

            W = (1.0 / 5.0) * (Wxxxx + Wyyyy + Wzzzz +
                                2.0 * Wxxyy + 2.0 * Wxxzz + 2.0 * Wyyzz)

            A = ((Wxxxx - W) * (Wxxxx - W) +
                 (Wyyyy - W) * (Wyyyy - W) +
                 (Wzzzz - W) * (Wzzzz - W) +
                 4.0 * (Wxxxy * Wxxxy + Wxxxz * Wxxxz +
                        Wxyyy * Wxyyy + Wyyyz * Wyyyz +
                        Wxzzz * Wxzzz + Wyzzz * Wyzzz) +
                 6.0 * ((Wxxyy - W / 3.0) * (Wxxyy - W / 3.0) +
                        (Wxxzz - W / 3.0) * (Wxxzz - W / 3.0) +
                        (Wyyzz - W / 3.0) * (Wyyzz - W / 3.0)) +
                 12.0 * (Wxxyz * Wxxyz + Wxyyz * Wxyyz + Wxyzz * Wxyzz))

            B = (Wxxxx * Wxxxx + Wyyyy * Wyyyy + Wzzzz * Wzzzz +
                 4.0 * (Wxxxy * Wxxxy + Wxxxz * Wxxxz +
                        Wxyyy * Wxyyy + Wyyyz * Wyyyz +
                        Wxzzz * Wxzzz + Wyzzz * Wyzzz) +
                 6.0 * (Wxxyy * Wxxyy + Wxxzz * Wxxzz + Wyyzz * Wyyzz) +
                 12.0 * (Wxxyz * Wxxyz + Wxyyz * Wxyyz + Wxyzz * Wxyzz))

            if B > 0.0 and W > 1e-8:
                KFA[v] = sqrt(A / B)
            else:
                KFA[v] = 0.0

    return np.asarray(KFA)
